#include "PGBasics.hh"
#include "GridSim.hh"
#include "SensorSim.hh"
#include "NeuralNet.hh"
#include "NeuralNetBatch.hh"
#include "BasicController.hh"    
#include "BinaryController.hh"
#include "FactoredController.hh"
#include "CyclicPolicyBias.hh"
#include "OLPomdp.hh"
#include "CRFController.hh"
#include "NACTransform.hh"
#include "test_traffic.hh"

using namespace std;
using namespace libpg;


int test_crf(int _argc,  char ** _argv)
{
	int seed, argc, app_id, ctl_id, use_NAC;
	char ** argv;
	bool auto_baseline;
	Simulator *simulator;
	Controller *controller;
	FactoredController::Controllers controllers;
	RLAlg *learner;	

	argv = new char * [255];		// maximum 255 arguments to the sampler initializer
	argc = 0;
	Utility::copy_arg(&argc, argv, _argc, _argv);

	Utility::get_arg(&seed, &argc, argv, "-sd", 2, time(NULL));
	srand (seed);

	parse_para(&argc, argv);

	// 1 for grid, 2  for sensor
	Utility::get_arg(&app_id, &argc, argv, "-app", 2, DFT_APP_ID);

	// 1 (default) for NN RL, 2 for CRF RL, 3 for mixture
	Utility::get_arg(&ctl_id, &argc, argv, "-ctl", 2, DFT_CTL_ID);

	// 0 (default) for not using NAC, 1 for using NAC
	Utility::get_arg(&use_NAC, &argc, argv, "-nac", 2, DFT_USE_NAC);
	
	Utility::get_arg(&auto_baseline, &argc, argv, "-abl", 0, DFT_AUTO_BASELINE);

	switch (app_id)
	{
		case 1:			// Grid
			simulator = new GridSim(&argc, argv);			
			break;
		case 2:			// Sensor.  2 and 3 are both for sensor, but 2 for chain CRF and 3 for circle CRF
		case 3:			// If used as NN, then the same.  
						// Though same application, deliberately made different for implementation convenience.
			simulator = new SensorSim(&argc, argv);
			break;
		default:
			cout<<"Unexpected simulator "<<app_id<<"."<<endl;
			exit(1);
	}

	switch (ctl_id)
	{
		case 1:			// NN RL
			if (app_id == 1)			// since there are 2 possible actions, so can use BinaryController
			{
				if (use_NAC)
				{
					for (seed = 0; seed < simulator->getAgents(); seed ++) 
						controllers.push_back(new NACTransform(new BinaryController(
						new NeuralNetBatch(simulator->getObsRows(), 1)), DFT_TD_DISCOUNT));				
				}
				else
				{
					for (seed = 0; seed < simulator->getAgents(); seed ++) 
						controllers.push_back(new BinaryController(new NeuralNet(simulator->getObsRows(), 1)));
				}
			}
			else if (app_id == 2 || app_id == 3)		// special since there are 3 possible actions, so must use BasicController
			{
				if (use_NAC)
				{
					for (seed = 0; seed < simulator->getAgents(); seed ++) 
						controllers.push_back(new NACTransform(new BasicController(
						new NeuralNetBatch(simulator->getObsRows(), 3)), DFT_TD_DISCOUNT));				
				}
				else
				{
					for (seed = 0; seed < simulator->getAgents(); seed ++) 
						controllers.push_back(new BasicController(new NeuralNet(simulator->getObsRows(), 3)));
				}
			}
			controller = new FactoredController(controllers, true, false);
			break;
		case 2:			// CRF RL
			controller = new CRFController(&argc, argv);
			if (use_NAC)
				controller = new NACTransform(controller, td_discount);
			break;
		default:
			cout<<"Unexpected controller "<<ctl_id<<"."<<endl;
			exit(1);
	}

	learner = new OLPomdp(controller, simulator, discount, step_size);
	learner->set_exit_para(&argc, argv);
	
    // This code should not be changed for true CRF code.

    //learner.read(PARAM_FILE);
    learner->saveBest(param_file);
	learner->setUseAutoBaseline(auto_baseline);	// for Grid, set to true. For Sensor net, set to false.

    time_t startTime = time(NULL);
    learner->learn(steps_per_epoch, max_rl_time, max_rl_steps);

    cout<<"Took "<<time(NULL) - startTime<<" secs\n";

    return 0;
}


bool check_traffic(int argc, char **argv)
{
	for(int i = 0; i < argc; i ++)
	{
		if (strstr(argv[i], "--"))
			return true;
	}
	return false;
}


int main(int argc, char** argv) 
{
	if(check_traffic(argc, argv))
		return test_traffic(argc, argv);

	return test_crf(argc, argv);
}


/*
//////////////////////////////////////////////////////////////////////////
//	Argument flags
// !- are command line arguments
// - are internal arguments for communication and arg passing
// If isn't must fill, then have have default values, with name DFT_XXX, or given by (default)

If command line contains '--' as argument, then treat as traffic problem

!-sd	Int		seed				(default) time(NULL)


==================================================
Problem specification

!-app	Int		app_id		dft: 1			1 for grid,  
											2 for sensor with supernodes
											3. sensor with CRF circle
											4. traffic network (offset scenario)

!-ctl	Int		ctl_id		dft: 1			1 for NN RL, 2 for CRF RL
!-nac	Int		use_NAC		dft: 0			0 for not using NAC, 1 for using NAC


==================================================
For RL
!-spe	Int		steps_per_epoch		
!-xs		Int		max_rl_steps		dft: 0
!-xt		Int		max_rl_time			dft: 0
!-ly		Int		nn_layers			dft: 2
!-xo		Int		approx_outputs		dft: 1
!-dc		Double	discount			dft: 0.6
!-ss1		Double	step_size			dft: 0.001
!-td		Double	td_discount			dft: 0.5
!-pf		char *	param_file (file name) dft: crf.params


==================================================
Exit control
!-ecw		Int		exit_check_window	dft: -1 (<0 means no check window)
!-ert		Double	exit_rel_tol		must fill if -ecw is set
!-eat		Double	exit_abs_tol		must fill if -ecw is set		



==================================================
For tree sampler:
!-xtt	Int		max_tree_time
!-xti	Int		max_tree_iter
-no		Int		num_obs_dim			must fill			// number of observation dimension
-ncl	Int		num_class			must fill			// dimension of label/action
-nef	Int		num_edge_fea		must fill			// number of edge feature
-nnxf	Int		num_node_extra_fea	dft: 0
!-et	Double	edge_fea_intensity	dft: 1
!-ss2	Double	stepSize[2]			dft: same as -ss			// For edge feature step size
!-ss3	Double	stepSize[3]			dft: same as -ss			// For extra node feature step size
!-xn	Double	maxNodeRand			dft: 0.1
1-xe	Double	maxEdgeRand			dft: 0.1
-na		Int		num_agent/global_num_node		must fill
!-ulr	bool	use_local_reward	dft: 0			
!-idy	int		min_delay			must fill		// for delay edge features
!-xdy	int		max_delay			must fill		// for delay edge features

!-erd	int		edge_reward_def		// edge reward definition	
			//	must fill if use_local_reward 	
			// 1: arithmetic mean of two end nodes (default) 
			// 2: geometric mean
			// 3: min of two end nodes
			// 4: max of two end nodes
			// 5: end node [2]
			// 6: end node [1]

!-efs:	Int		edge_fea_sel		// edge feature selector
		def: 0
		& 0x1: constant edge bias (pair of current), ONE feature just compare same or not
		& 0x2: constant edge bias (pair of current), 16 features for all possible combinations
		
		& 0x4: left [1] past, self [2] current, ONE feature just compare same or not
		& 0x8: left [1] past, self current, 16 features for all possible combinations

		& 0x16: left [1] current, self [2] past, ONE feature just compare same or not
		& 0x32: left [1] current, self [2] past, 16 features for all possible combinations

// For cycle CRF (cycle bias)
-cyt	Int		cycleTime								// if > 0, then use cycleTime
-bias	Double	bias				must fill if cycleTime > 0

// For cross CRF
-ncu	Int		num_cross_upper		must fill
-chi	Int		cross_hori_id		must fill

// For debug
-v		Int		verbosity			0
		Control of verbosity level (output of intermediate results)
		1.  Display all the trees and init_tree
		2.  Display the RB estimates of nodes after all iterations of tree MCMC sampling
		4.  Display the RB estimates of edges after all iterations of tree MCMC sampling
		8.  Display the node and edge potentials at the outset
		16. Display number of iterations if max_time is used



==================================================
Simulator specific:

// For GridSim
-nr		Int		num_row				In GridSim dft: 3, in CRFGrid must fill
-nc		Int		num_col				In GridSim dft: 3, in CRFGrid must fill
-pg		Double	packetGenProb		0.5
-bs		Int		bufferSize			10

// For SensorSim
-nc		Int		num_cell			NUM_TARGET
-nt		Int		num_target			NUM_TARGET
-xy		Int		max_cycle			0					// 0 for infinity


// For CRFController and Tree sampler


and there is a pre-process directive in CRFBasic
#define MY_DEBUG


*/

