#include"PGBasics.hh"
#include"Sampler.hh"
#include"ManyToOneController.hh"


using namespace std;
namespace libpg {

/**
 * @param  approximators vector of Approximator* 
 * @param  if true, expect as many colums in the observations as there are approximators
 * and each controller only gets its col as an observation.
 * If false, all approximators get the same observation.
 */
ManyToOneController::ManyToOneController(Approximators approximators, bool splitObs, bool localRewards) {

    this->approximators = approximators;
    this->splitObs = splitObs;
    this->localRewards = localRewards;
    dummyDist.resize(1);
    dist.resize(approximators.size());

}


/**
 * Where the real coding is done to implement softmax
 */
void ManyToOneController::getAction(Observation& obs, Vector& action,  bool computeGrad) {
    
    double oneNorm = 0.0;

    // Can only have one winning action here!
    assert(action.size() == 1);

    if (splitObs) {
	// If we have a col for each controller, they get their own
	// observation.
	assert(obs.getFeatures().size2() == approximators.size());
    } else {
	assert(obs.getFeatures().size2() == 1);
    }

    // Loop over possible actions, each is its own approximator
    for (unsigned int i = 0; i < approximators.size(); i++) {
	if (!obs.getEligible()(i)) { // This action not eligible.
	    dist[i] = 0.0;
	    continue;
	}
	if (splitObs) obs.setAgent(i);
	approximators[i]->doApprox(obs, dummyDist);
	dist[i] = exp(dummyDist[0]);
	oneNorm += dist[i];
    }

    // Pick an action
    int sampledAct  = Sampler::discrete(dist, oneNorm);
    action[0] = sampledAct;
    
    if (computeGrad) {
	dist *= -1.0/oneNorm;
	dist[sampledAct] += 1.0;
	// Gradient will be added directly to trace for each approximator

	for (unsigned int i = 0; i < approximators.size(); i++) {
	    if (!obs.getEligible()(i)) continue; // This action not eligible.
	    if (splitObs) obs.setAgent(i);
	    dummyDist(0) = dist[i];
	    approximators[i]->feedbackGrad(obs, dummyDist);
	}
    }
}


/**
 * Pick the controller that has the highest probability of being chosen.
 * @param  observation vector
 * @param  empty vector length 1 to put the action in.
 */
void ManyToOneController::getMostProbAction(Observation& obs, Vector& action) {

    // Can only have one winning action here!
    assert(action.size() == 1);

    if (splitObs) {
	// If we have a col for each controller, they get their own
	// observation.
	assert(obs.getFeatures().size2() == approximators.size());
    } else {
	assert(obs.getFeatures().size2() == 1);
    }

    // Loop over possible actions, each is its own approximator
    for (unsigned int i = 0; i < approximators.size(); i++) {
	if (!obs.getEligible()(i)) { // This action not eligible.
	    dist[i] = 0.0;
	    continue;
	}
	if (splitObs) obs.setAgent(i);
	approximators[i]->doApprox(obs, dummyDist);
	dist[i] = exp(dummyDist[0]);
    }

    // Pick an action
    action[0] = index_norm_inf(dist);
    
}


int ManyToOneController::getInputDim() {
    return approximators[0]->getInputDim();
}


/**
 *
 * The remaining methods just pass onto equiv Approximator
 * functions. We have to do it in this ugly way because more advanced
 * multiple agent approximators will need to override these functions to
 * update all agents in turn.
 *
 */

void ManyToOneController::discountTrace() {

    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->discountTrace();
    }

}


void ManyToOneController::setDiscount(double discount) {
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->setDiscount(discount);
    }
}


void ManyToOneController::accumulateGrad(Vector& rewards, Observation& newObs) {

    double reward;

    assert((localRewards && rewards.size()==approximators.size()) || 
	   (!localRewards && rewards.size()==1));
    
    for (uint i = 0; i < approximators.size(); i++) {
	reward =  localRewards?rewards[i]:rewards[0];
	if (reward != 0.0) approximators[i]->accumulateGrad(reward, newObs);	
    }
}


void ManyToOneController::resetGrad() {

    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->resetGrad();
    }

}


void ManyToOneController::instantStep(Vector& rewards) {

    double reward;
    assert((localRewards && rewards.size()==approximators.size()) || 
	   (!localRewards && rewards.size()==1));

    for (uint i = 0; i < approximators.size(); i++) {
	reward =  localRewards?rewards[i]:rewards[0];
	if (reward != 0.0) approximators[i]->instantStep(reward);
    }
}


void ManyToOneController::computeDirection(int steps) { 
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->computeDirection(steps);
    }
}


void ManyToOneController::batchStep() { 
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->batchStep();
    }
}

void ManyToOneController::setStepSize(double stepSize) { 
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->setStepSize(stepSize);
    }
}


void ManyToOneController::resetTrace() { 
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->resetTrace();
    }
}


void ManyToOneController::resetParams() { 
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->resetParams();
    }
}


void ManyToOneController::randomizeParams(double maxRand) { 
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->randomizeParams(maxRand);
    }
}


double ManyToOneController::getMaxParam() {
    double maxi = 0.0;
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	maxi = std::max(maxi, (*i)->getMaxParam());
    }
    return maxi;
}


int ManyToOneController::getNumParams() {
    int p = 0;
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	p += (*i)->getNumParams();
    }
    return p;
}


void ManyToOneController::write(ostream& o) {
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->write(o);
    }
}
  

void ManyToOneController::read(istream& o) {
    Approximators::iterator i;
    for (i = approximators.begin(); i != approximators.end(); i++) {
	(*i)->read(o);
    }
}  
}
