/**
 * $Id: NACTransform.cc 29 2006-12-19 02:16:20Z daa $
 */

#include"PGBasics.hh"
#include"NeuralNet.hh"
#include"NeuralNetBatch.hh"
#include"NACTransform.hh"
#include "CRFController.hh"

using namespace std;
namespace libpg {

/**
 * Constructor. Set up additional vectors and matrices needed to
 * implement natural actor critic.
 */
NACTransform::NACTransform(Controller* controller, double tdDiscount) 
    : TransformController(controller) 
{
    this->tdDiscount = tdDiscount;

	if (NACTransform::print_nac_info)  //only print once, if there are 16 factored controllers.
	{
		NACTransform::print_nac_info = false;
		cout<<"Using NAC, td discount = "<<tdDiscount<<endl;
	}

    // We turn int observations into unit vectors for first term
    // Second term comes from number of parameters.
    nacDimension = controller->getInputDim() + controller->getNumParams();

    I.resize(nacDimension, nacDimension);
    I.clear();

    //A.resize(nacDimension, nacDimension);
    AInverse.resize(nacDimension, nacDimension);
    AInverse.clear();
 
    nacTD.resize(nacDimension);
    nacTrace.resize(nacDimension);
    nacTrace.clear();
    nacGrad.resize(nacDimension);
    nacGrad.clear();
    nacDirection.resize(nacDimension);
    nacDirection.clear();
    nacParamUpdate.resize(controller->getNumParams());
    nacParamUpdate.clear();
    I.assign(identity_matrix<double>(nacDimension));

	vec_lastReward.resize(controller->get_num_agent());
         
    instantStepRequired = false;
	
	if (controller->distribute_reward())
	{
		CRFController * crf_controller = (static_cast <CRFController *> (controller));
		rewards = new double [crf_controller->get_num_agent() + crf_controller->get_num_edge() + 1];
	}
	else
		rewards = new double [2];
}

 
/**
 * NAC delays it's discounting till the end of a updateStatistics()
 */
void NACTransform::discountTrace() 
{
}


/**
 * Just a wrapper that intercepts the new observation and completes
 * the instant step from the last round if necessary.
 */
void NACTransform::getAction(Observation& obs, Vector& action, bool computeGrad) 
{
	int cur_upper, cur_index, cur_item;
	if (controller->is_independent())
		assert(obs.getFeatures().size1() == (size_t)getInputDim());	
    else
		assert(obs.getFeatures().size1() * obs.getFeatures().size2() == (size_t)getInputDim());	

    
    if (instantStepRequired) 
	{
		if (controller->distribute_reward())		// use CRF AND local reward
			dist_rwd_delayedInstantStep(vec_lastReward, obs);		
		else										// CRF with global reward or Independent
			delayedInstantStep(lastReward, obs);
	}

	// Since for NAC, the trace in the controller is not important (NAC maintains its own NACTrace)
	// The reduce will use the one-step (most recent) trace (the gradient in step 6 and 11)
    controller->resetTrace();
	
    controller->getAction(obs, action, computeGrad);

	if (computeGrad) 
	{
		// Grab the most recent (one-step) trace
		controller->reduce(nacTrace, Approximator::TRACES);

		// Add in the observed feature (basis vector) for the current
		// state into the bottom part of the nacTrace vector.
		// comment this line out if want to use natural gradient

		if (controller->is_independent())		// use distributed reward
		{	
			project(nacTrace, range(controller->getNumParams(), nacDimension)) 
					+= column(obs.getFeatures(), obs.getAgent());
		}
		else
		{
			cur_index = controller->getNumParams();
			for (cur_item = 0; cur_item < controller->get_num_agent(); cur_item ++)
			{
				cur_upper = cur_index + obs.getFeatures().size1();
				project(nacTrace, range(cur_index, cur_upper)) 
						+= column(obs.getFeatures(), cur_item);
				cur_index = cur_upper;
			}
		}

		// Make a copy of the last observation.
		lastObs = obs;
	}
}


void NACTransform::resetTrace() 
{
    nacTrace.clear();
    AInverse.assign(I); 
}


void NACTransform::resetGrad() 
{
    nacGrad.clear();
    nacTrace.clear();  
    AInverse.assign(I);  
}



/**
 * In the case of NAC this is a bit of a hack. The update requires the next 
 * observation to work, so we'll store the fact that an update is required
 * and do it in the next feedback() phase. This effectively means you end
 * up with steps-1 updates, but oh well...
 */
void NACTransform::instantStep(Vector& reward) 
{
    if (reward.size() > 1 && !controller->distribute_reward()) 
		throw runtime_error("NACTransform::instantStep() requires scalar reward for global reward\n");
	
    instantStepRequired = true;
    if (controller->distribute_reward())
		vec_lastReward.assign(reward);
    else
		lastReward = reward[0];
}


void NACTransform::delayedInstantStep(double reward, Observation& newObs)
{
    assert(instantStepRequired);

    assert(controller->is_independent() && newObs.getFeatures().size1() == (size_t)getInputDim()
		|| !controller->is_independent() 
			&& newObs.getFeatures().size1() * newObs.getFeatures().size2() == (size_t)getInputDim());

    nacGrad.clear();

	updateStatistics(reward, newObs);

    // Divide gradient by A, effectively applying second order statistics
	// Step 16 
    axpy_prod(AInverse, nacGrad, nacDirection, true); 
    
    // Update the parameters. step 17
    batchStep();

    instantStepRequired = false;
}


void NACTransform::dist_rwd_delayedInstantStep(Vector & reward, Observation& newObs) 
{
    assert(instantStepRequired);

    assert(newObs.getFeatures().size1() * newObs.getFeatures().size2() == (size_t)getInputDim());

    nacGrad.clear();
 
	dist_rwd_updateStatistics(reward, newObs);

    // Divide gradient by A, effectively applying second order statistics
    axpy_prod(AInverse, nacGrad, nacDirection, true); 
    
    // Update the parameters. step 17
    batchStep();

    instantStepRequired = false;
}



void NACTransform::accumulateGrad(double reward, Observation& newObs) 
{
	NYI
    updateStatistics(reward, newObs);
}


// update nacGrad, nacTD, A_t^{-1}, 
void NACTransform::updateStatistics(double reward, Observation& newObs) 
{	
	int cur_index, cur_item, cur_upper;

    assert(controller->is_independent() && newObs.getFeatures().size1() == (size_t)getInputDim()
		|| !controller->is_independent() 
			&& newObs.getFeatures().size1() * newObs.getFeatures().size2() == (size_t)getInputDim());

    // Construct the nacTD vector
    nacTD.clear();

    controller->reduce(nacTD, Approximator::TRACES);		// for the gradient part of step 11 (w_t)


	// The next part is to implement the x_t - \gamma * x_{t+1} in step 11
	if (controller->is_independent())
	{
		// Add in the observed feature (basis vector) for the current
		// state into the bottom part of the nacTD vector.
		// comment this line out if want to use natural gradient

		// Also subtract off the future observation * discount
		// comment this line out if want to use natural gradient

		project(nacTD, range(controller->getNumParams(), nacDimension)) = 
				column(lastObs.getFeatures(), lastObs.getAgent())
				- tdDiscount * column(newObs.getFeatures(), newObs.getAgent());				
	}
	else
	{
		cur_index = controller->getNumParams();

		for (cur_item = 0; cur_item < controller->get_num_agent(); cur_item ++)
		{
			cur_upper = cur_index + lastObs.getFeatures().size1();
			project(nacTD, range(cur_index, cur_upper)) 
				= column(lastObs.getFeatures(), cur_item);
			
			project(nacTD, range(cur_index, cur_upper)) 
				-= tdDiscount * column(newObs.getFeatures(), cur_item);

			cur_index = cur_upper;
		}
	}
	


    // OK, we have computed the TD part of the vector. Now do outer
    // product with trace to update A Although in fact, since what we
    // want in the end is the inverse of A, lets update that directly.

    double alpha = 1.0 - 1.0/(lastObs.getSteps()+1);
    assert(alpha > 0.0);
    
    //A += (1.0-alpha)*(outer_prod(nacTD, nacTD)+1e-2*I);//natural gradient ggT
    //A += (1.0-alpha)*(outer_prod(nacTrace, nacTD)+1e-2*I);
    //A += outer_prod(nacTrace, nacTD); //it is easily to blow up at some stage
    //A += outer_prod(nacTrace, nacTD)+1e-2*identity_matrix<double>(nacDimension);
    //cout<<"A"<<A<<endl;


    // Update inverse of A
    AInverse /= alpha; // same as multiplying A by alpha)
    nacTD *= (1.0 - alpha);		// nacTD will only be used in the following command, so just change it
    UBlasExtras::updateInverseMatrix(AInverse, nacTrace, nacTD);	// update the inverse of A
	
	
    // Update normal (GPOMDP style) gradient
	// since nacGrad is cleared to 0 in delayedInstantStep, so += is actually =
	// step 9 in pseudo-code
	nacGrad += reward * nacTrace;		

    // Update the trace out of order (normal discountTrace command ignored).
	// \beta * e_{t-1} in step 6
    nacTrace *= discount;
}



void NACTransform::dist_rwd_updateStatistics(Vector & reward, Observation& newObs) 
{	
	int cur_index, cur_reward_index, cur_item, cur_length, cur_upper;
	double cur_reward;
    
    assert(newObs.getFeatures().size1() * newObs.getFeatures().size2() == (size_t)getInputDim());

	CRFController * crf_controller = static_cast <CRFController *> (controller);

    // Construct the nacTD vector 
    nacTD.clear();

    controller->reduce(nacTD, Approximator::TRACES);

	cur_index = controller->getNumParams();
	for (cur_item = 0; cur_item < crf_controller->get_num_agent(); cur_item ++)
	{
		// Add in the observed feature (basis vector) for the current
		// state into the bottom part of the nacTD vector.
		// comment this line out if want to use natural gradient
		cur_upper = cur_index + lastObs.getFeatures().size1();

		// Also subtract off the future observation * discount
		project(nacTD, range(cur_index, cur_upper)) = 
			column(lastObs.getFeatures(), cur_item)
			- tdDiscount * column(newObs.getFeatures(), cur_item);

		cur_index = cur_upper;
	}

	// OK, we have computed the TD part of the vector. Now do outer
	// product with trace to update A Although in fact, since what we
	// want in the end is the inverse of A, lets update that directly.

    double alpha = 1.0 - 1.0/(lastObs.getSteps()+1);
    assert(alpha > 0.0);
    
    //A += (1.0-alpha)*(outer_prod(nacTD, nacTD)+1e-2*I);//natural gradient ggT
    //A += (1.0-alpha)*(outer_prod(nacTrace, nacTD)+1e-2*I);
    //A += outer_prod(nacTrace, nacTD); //it is easily to blow up at some stage
    //A += outer_prod(nacTrace, nacTD)+1e-2*identity_matrix<double>(nacDimension);
    //cout<<"A"<<A<<endl;


    // Update inverse of A
    AInverse /= alpha;		// same as multiplying A by alpha)
    nacTD *= 1.0 - alpha;
    UBlasExtras::updateInverseMatrix(AInverse, nacTrace, nacTD);

	// Update normal (GPOMDP style) gradient
	for (cur_index = 0; cur_index < crf_controller->get_num_agent(); cur_index ++)
		rewards[cur_index + 1] = reward[cur_index];

	crf_controller->fill_edge_reward(rewards);

	cur_index = 0;
	cur_length = newObs.getFeatures().size1() * crf_controller->get_num_class();
	cur_reward_index = 1;
	
	for (cur_item = 1; cur_item <= crf_controller->get_num_agent(); cur_item ++)
	{
		cur_reward = rewards[cur_reward_index ++];
		cur_upper = cur_index + cur_length;
 
		project(nacGrad, range(cur_index, cur_upper)) 
			+= cur_reward * project(nacTrace, range(cur_index, cur_upper));

		cur_index = cur_upper;
	}

	if (crf_controller->get_num_node_extra_fea())
	{
		cur_reward_index = 1;
		cur_length = crf_controller->get_num_node_extra_fea();
		for (cur_item = 1; cur_item <= crf_controller->get_num_agent(); cur_item ++)
		{
			cur_reward = rewards[cur_reward_index ++];
			cur_upper = cur_index + cur_length;
 
			project(nacGrad, range(cur_index, cur_upper)) 
				+= cur_reward * project(nacTrace, range(cur_index, cur_upper));

			cur_index = cur_upper;
		}
	}

	for (cur_item = 1; cur_item <= crf_controller->get_num_edge(); cur_item ++)
	{
		cur_reward = rewards[cur_reward_index ++];
		cur_upper = cur_index + crf_controller->get_num_edge_fea();
 
		project(nacGrad, range(cur_index, cur_upper)) 
			+= cur_reward * project(nacTrace, range(cur_index, cur_upper));

		cur_index = cur_upper;
	}

	cur_reward_index = 1;
	cur_length = newObs.getFeatures().size1();
	for (cur_item = 1; cur_item <= crf_controller->get_num_agent(); cur_item ++)
	{
		cur_reward = rewards[cur_reward_index ++];
		cur_upper = cur_index + cur_length;

		project(nacGrad, range(cur_index, cur_upper)) 
			+= cur_reward * project(nacTrace, range(cur_index, cur_upper));

		cur_index = cur_upper;
	}
	
    // Update the trace out of order (normal discountTrace command ignored).
    nacTrace *= discount;
}


/**
 * Normalize the gradient and multiply by the inverse of the
 * A matrix, doing the magic step of natural actor-critic in one
 * go.
 * Could use Sherman-Morrison:
 * (A + u^T v)^-1 = A^-1 - (A^-1 u)^T (v A^-1) / (1 + v A^-1 u)
 * @param  how many steps did we run to compute grad
 */
void NACTransform::computeDirection(int steps) 
{    
    nacGrad /= steps;

    axpy_prod(AInverse, nacGrad, nacDirection, true); 
    //cout<<"direction="<<nacDirection<<endl;

}


void NACTransform::batchStep() 
{ 
	// get d_t from the matrix-vector multiplication in step 16
    nacParamUpdate.assign(project(nacDirection, range(0, controller->getNumParams())));

    // Loop over each layer of the neural network
	// distribute the d_t to the space of trace in the controller (borrowing)
    controller->scatter(nacParamUpdate, Approximator::GRADS);

	// update the parameters by using the information in the trace space (step 17)
	// add to the parameters the trace value * step_size
    controller->batchStep();
    
}


void NACTransform::reduce(Vector& v, Approximator::StatsEnum s) 
{
    switch (s) 
	{
		case Approximator::TRACES:		// Deliberate fall through, actually only this is used
		case Approximator::PARAMS: 
			controller->reduce(v, s);
			break;
		case Approximator::GRADS: 
			assert(v.size() == (size_t)getNumParams());
			v.assign(project(nacDirection, range(0, getNumParams())));
			break;
		default:
			NYI
    }
}


// almost not used
void NACTransform::scatter(Vector& v, Approximator::StatsEnum s) 
{     
    switch (s) 
	{
		case Approximator::PARAMS: // deliberate fall through
		case Approximator::TRACES:	// only use this case
			controller->scatter(v, s);
			break;
		case Approximator::GRADS:
			assert(v.size() == (size_t)getNumParams());
			project(nacDirection, range(0, getNumParams())).assign(v);
			break;
		default:
			throw runtime_error("NACTransform::scatter() unknown type\n");
    }

}


/**
 * Takes a local copy of discount and passes it back down to next controller
 */
void NACTransform::setDiscount(double discount) 
{
    this->discount = discount;
    controller->setDiscount(discount);
}


char * NACTransform::get_diagnosis(int argc, char ** argv)
{
	return controller->get_diagnosis(argc, argv);
}



bool NACTransform::print_nac_info = true;

}
