#include"PGBasics.hh"
#include"BasicController.hh"
#include"SARSAController.hh"
#include"Sampler.hh"

namespace libpg {


/**
 * SARSA controller constructor
 * @param approx parameterised distribution we'll be using today
 * @param tdDicount temporal difference discount
 */
SARSAController::SARSAController(Approximator* approx, double tdDiscount) : BasicController(approx) {
    this->tdDiscount = tdDiscount;
    lastRew = 0.0;
    lastActVal = 0.0;
}


/**
 * Select the next action based on softmax with Gibbs (or Boltzmann) distribution.
 * @param obs observation vector
 * @param action single position vector used to return the selected action
 * @param computeGrad flag used to enable learning from experience
 */
void SARSAController::getAction(Observation& obs, Vector& action,  bool computeGrad) {
    
    int sampledAct;
    double oneNorm = 0;
    Vector actionValues;

    assert(action.size() == 1);

    // Learn by experience?
    this->computeGrad = computeGrad;

    // Get state-action values for this state
    approx->doApprox(obs, dist);

    // Backup dist to get sampled action's state-action value
    actionValues = dist;
    
    // Element wise exponentiation and 1-norm, first step of softmax
    Vector::iterator i;
    for (i = dist.begin(); i != dist.end(); i++) {
	*i = exp(*i);
	oneNorm += *i;
    }

    // Pick an action
    sampledAct = Sampler::discrete(dist, oneNorm);
    assert(sampledAct < approx->getOutputDim());
    action[0] = sampledAct;

    // Temporal difference error calculation
    error = lastRew + tdDiscount * actionValues(sampledAct) - lastActVal;
    
    // Recall last action, observation and action value
    lastAct = sampledAct;
    lastObs = obs;
    lastActVal = actionValues(sampledAct);
}


/**
 * Updates the eligibility traces
 */
void SARSAController::discountTrace() {
    if (computeGrad) {
        approx->discountTrace();
	for (Vector::iterator i = dist.begin(); i != dist.end(); i++)
	    *i = 0.0;
        dist[lastAct] = 1.0;
        approx->feedbackGrad(lastObs, dist);
    }
}


/**
 * Sets the discount factor
 */
void SARSAController::setDiscount(double discount) {
    this->discount = discount;
    approx->setDiscount(tdDiscount * discount);
}


/**
 * Updates the parameter vector using eligibility traces
 * @param reward received on previous interacion with
 * environment
 */
void SARSAController::instantStep(Vector& reward) {
    if (computeGrad) {
        lastRew = reward[0];
        approx->instantStep(error);
    }
}


}
