
// $Id$

#include "PGBasics.hh"
#include "GridSim.hh"
#include "Utility.hh"

// Defaults
#define GRID_SIZE_ROW 3
#define GRID_SIZE_COL 3
#define PACKET_GEN_PROB 0.5
#define BUFFER_SIZE 10
#define OBS_DIM 2

#define DBG_POL false

using namespace std;
using namespace libpg;


/**
 * @param gridSize how many paths in for each direction, e.g., square
 * root of number of agents.
 * @param frequency with which packets arrive. 
 * @param bufferSize new packets will be dropped if more than this
 * packets waiting in any queue
 */
GridSim::GridSim(int *argc, char ** argv) 
{
	int i;
	Utility::get_arg(&num_row, argc, argv, "-nr", 0, GRID_SIZE_ROW);
	Utility::get_arg(&num_col, argc, argv, "-nc", 0, GRID_SIZE_COL);
	Utility::get_arg(&packetGenProb, argc, argv, "-pg", 0, PACKET_GEN_PROB);
	Utility::get_arg(&bufferSize, argc, argv, "-bs", 0, BUFFER_SIZE);

    nodes = num_row * num_col;
	Utility::get_arg(&ctl_id, argc, argv, "-ctl", 1, 0);			// 1 for NN, 2 for CRF grid

	if (ctl_id == 2)
	{
		Utility::add_arg(argc, argv, nodes, "-na");
		Utility::add_arg(argc, argv, 1, "-nef");
		Utility::add_arg(argc, argv, 2, "-ncl");
		Utility::add_arg(argc, argv, 2, "-no");
		Utility::add_arg(argc, argv, 0, "-nnxf");
	}

    // Init queues
    row_queue = new int[num_row];
    col_queue = new int[num_col];
    for (i = 0; i < num_row; i ++) 
		row_queue[i] = 0;
    for (i = 0; i < num_col; i ++) 
		col_queue[i] = 0;

	count_act1 = count_act2 = 0;
    cout <<"Initialized GridSim. Max average reward = "<<(num_row + num_col) * packetGenProb << endl;
}


GridSim::~GridSim() 
{
	delete [] row_queue;
	delete [] col_queue;
}


inline int GridSim::getObsRows() { return OBS_DIM; }

inline int GridSim::getObsCols() { return nodes; }

inline int GridSim::getAgents() { return nodes; }

inline int GridSim::getActionDim() { return nodes; }

inline double GridSim::get_max_reward() { return (num_row + num_col) * packetGenProb; }


/**
 * Global rewards 
 */
inline int GridSim::getRewardDim() { return 1; }

/**
 * Just return the next reward set in doAction
 */
void GridSim::getReward(Vector& rewards) {

    assert(rewards.size() == 1);
    rewards[0] = nextReward;

}


/**
 * Two observations per node giving normalized number of waiting packets. 
 * @param obs observation structure that should be correct size already
 */
void GridSim::getObservation(Observation& obs) {

	int cur_index, cur_row, cur_col;

    assert((int)obs.getFeatures().size1() == OBS_DIM);
    assert((int)obs.getFeatures().size2() == nodes);

	cur_index = 0;
    for (cur_row = 0; cur_row < num_row; cur_row ++) {
		for (cur_col = 0; cur_col < num_col; cur_col ++) {
			obs.getFeatures()(0, cur_index) = row_queue[cur_row] / (double) bufferSize;
			obs.getFeatures()(1, cur_index ++) = col_queue[cur_col] / (double) bufferSize;
		}
    }

}


/**
 * Do the suggested action, passing through any packets if possible,
 * then add in new packets
 */
int GridSim::doAction(Vector& action) 
{
	int cur_row, cur_col, cur_index, cur_head;

    nextReward = 0;
    assert((int)action.size() == nodes);

    if (DBG_POL) cout<<"Col: ";

	// Look at each column, testing for a clear path
    bool clear = true;
    bool clearCol = false;

	for (int i = 0; i < nodes; i ++)
	{
		if(action[i])
			count_act2 ++;
		else
			count_act1 ++;
	}

	cur_head = 0;
    for (cur_col = 0; cur_col < num_col; cur_col ++) 
	{
		clear = true;
		cur_index = cur_head;

		for (cur_row = 0; cur_row < num_row; cur_row ++) 
		{
			if (int(action[cur_index] + 0.5) != VERTICAL) 
			{
				clear=false;
				break;
			}
			cur_index += num_col;
		}
		if (clear)
		{
			// There's a clear column. All waiting traffic gets
			// through and contributes to the reward
			nextReward += col_queue[cur_col];
			col_queue[cur_col] = 0;
			if (DBG_POL)  cout<<cur_col<<' ';
			clearCol = true;
		}		
		cur_head ++;
    }


    // Look at each row, testing for a clear path
    // actually, if any column was clear, this code is redundant.
	if (!clearCol)
	{
		cur_head = 0;
		for (cur_row = 0; cur_row < num_row; cur_row ++) 
		{
			clear = true;
			cur_index = cur_head;

			for (cur_col = 0; cur_col < num_col; cur_col ++) 
			{
				if (int(action[cur_index] + 0.5) != HORIZONTAL) 
				{
					clear=false;
					break;
				}
				cur_index ++;
			}
			if (clear)
			{
				// There's a clear column. All waiting traffic gets
				// through and contributes to the reward
				nextReward += row_queue[cur_row];
				row_queue[cur_row] = 0;
				if (DBG_POL)  cout<<cur_row<<' ';
			}		
			cur_head += num_col;
		}
	}

    // Now introduce new traffic
    for (cur_row = 0; cur_row < num_row; cur_row ++) 
	{
		if (row_queue[cur_row] < bufferSize && random() / (double)RAND_MAX <= packetGenProb) 
			row_queue[cur_row] ++;
	}

	for (cur_col = 0; cur_col < num_col; cur_col ++) 
	{
		if (col_queue[cur_col] < bufferSize && random() / (double)RAND_MAX <= packetGenProb) 
			col_queue[cur_col] ++;
	}

	if (DBG_POL) 
	{
		cout<<"Row Queue:\t";
		for (cur_row = 0; cur_row < num_row; cur_row ++)
			cout << cur_row + 1 << ":" << row_queue[cur_row] << "\t";

		cout<<"\nColumn Queue\t";
		for (cur_col = 0; cur_col < num_col; cur_col ++)
			cout << cur_col + 1 << ":" << col_queue[cur_col] << "\t";

		cout<<endl;
    }

    return 0;
}


char * GridSim::get_diagnosis()
{
	sprintf(diagnosis_msg, ": %g", count_act1 / (double) count_act2);
	count_act1 = count_act2 = 0;

	return diagnosis_msg;
}
