#include "PGBasics.hh"
#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"

#include "iostream"

using namespace std;
using namespace libpg;

void Intersection::reset() { // removes cars and memories of cars, resets statistics
  reset_statistics();
  for(int i = 0; i < DIRECTIONS; i++) {
    cars_at_segment[i].clear();
    cars_at_segment[i].resize(Ps::pROAD_LENGTH + 1, 0); 
    std::vector<int>::iterator it;
    for(int j = 0; j < QUEUES; j++) {
      cars_queued[i][j] = 0;
      //cars_queued_set[i][j].clear();
      cars_queued_list[i][j].clear();
      blocked[i][j] = false;
      for(int p = 0; p < MAX_CYCLE_LENGTH; p++) 
	detector_counts[p][i][j] = 0;
    }
    //detector_pointer = -1;
  }
}

Intersection::Intersection(int x_coordinate, int y_coordinate) {
  // NOTE: incoming and outgoing neighbouring intersections have to be set
  // by the user after creation, as well as the length and capacity of the incoming roads
  // and dummy flag

  //MAX_CYCLE_LENGTH = Ps::pMAX_PHASE_DURATION * PHASES;
  //MAX_CYCLE_LENGTH = RPC_CYCLE_LENGTH;
  MAX_CYCLE_LENGTH = Ps::pMAX_CYCLE_LENGTH;
  RPC_CYCLE_LENGTH = Ps::pRPC_CYCLE_LENGTH;
  //RPC_CYCLE_LENGTH = 16; // only for RandomPhaseController
  if(!(MAX_CYCLE_LENGTH >= RPC_CYCLE_LENGTH)) {
    cout << "MAX_CYCLE_LENGTH must be greater or equal to RPC_CYCLE_LENGTH!" << endl;
    exit(1);
  }

  //cout << "! intersection (" << x_coordinate<< "," << y_coordinate 
  //   <<") produced !" << endl;  

   _VERBOSE_INTERSECTION = false;
  //_VERBOSE_INTERSECTION = (x_coordinate == 5 && y_coordinate == 2);

  coordinates[X_COORDINATE] = x_coordinate;
  coordinates[Y_COORDINATE] = y_coordinate;
  dummy = false;
  save_history = false;
  for(int i = 0; i < DIRECTIONS; i++) {
    incoming[i] = NULL;
    outgoing[i] = NULL;
    length_incoming[i] = -1;
    capacity_incoming[i] = -1;
  }
  detector_counts = new int[MAX_CYCLE_LENGTH][DIRECTIONS][QUEUES];
  summed_phase_counts = new int[RPC_CYCLE_LENGTH][PHASES];
  detector_pointer = -1;
  starting_phase = current_phase = (Phase) 0; // default value
  phase_duration = -1; //0
  cycle_duration = -1; //0
  waiting_cars_last_step = 0;
  
  for(int r = 0; r < RPC_CYCLE_LENGTH; r++) {
    for(int i = 0; i < PHASES; i++) {
      avg_phase_percent[i] = 0;
      max_car_throughput_last_cycle[i] = 0;
      phase_counts[i] = 0;
      summed_phase_counts[r][i] = 0;
      waiting_times_last_step[i] = 0;
      cars_passed_last_step[i] = 0;
    }
  }
  //phase_counts[0] = 1; // take this line out it it doesn't work...
  reset(); 
}

void Intersection::delete_neighbour(Direction dir) {
  //cout << "in neighbour ";
  //print_id();
  //cout << ", deleting direction" << dir << ": ";
  //outgoing[dir ]->print_id();
  //cout << endl;
  outgoing[dir] = NULL;
  incoming[dir] = NULL;
  length_incoming[dir] = -1;
}

Intersection::~Intersection() {
  //cout << "deleting ";
  //print_id();
  //cout << endl;
  for(int i = 0; i < DIRECTIONS; i++) {
    if(incoming[i] != NULL) {
      //cout << "sending neighbour " << i << " the delete message" << endl;
      //cout << "directions " << DIRECTIONS << " " << i << " " << ((i + 2) % DIRECTIONS) << " " <<
      //((Direction) ((i + 2) % DIRECTIONS)) << endl;
      incoming[i]->delete_neighbour( (Direction) ((i + 2) % DIRECTIONS) );  
      //incoming[i]->outgoing[((Direction) ((i + 2) % DIRECTIONS))] = NULL;
    }
  }
  delete detector_counts;
}

void Intersection::print_id() {
  cout << "Intersection (" << this->coordinates[X_COORDINATE] << "," 
       << this->coordinates[Y_COORDINATE] << ")";
} 

int Intersection::get_missing_phase() { // returns -1 if no phase is missing
  int missing_phases = 0;
  int possible_next = -1;
  for(int i = 0; i < PHASES; i++) {
    if(phase_counts[i] == 0) {
      //cout << "phase "<<i<<" has not been on" << endl;
      if(possible_next == -1) {
	possible_next = i;
      }
      missing_phases++;
    }
  }
  return possible_next;
}

void Intersection::print_connections() {
  print_id();
  cout << " is connected to (NESW):" << endl; 
  for(int i = 0; i < DIRECTIONS; i++) {
    if(outgoing[i] != 0)
      outgoing[i]->print_id();
    else
      cout << "-";
    cout << endl;
  }
  cout << endl;
}

void Intersection::set_history_saving(string filename) {
  print_id();
  cout << "saving history" << endl;
  /*
  string filen = ((string) filename) + "_intersection_" +
    ((string) ('0' + coordinates[X_COORDINATE])) + 
    ((string) ('0' + coordinates[Y_COORDINATE]));
  const char* fn = filen.c_str();
  */
  history_file = new ofstream(filename.c_str());  
  if (!history_file->is_open()) {
    cout<<"!!! Failed to open "<<endl;
  }
    else
      save_history = true;
}

void Intersection::unset_history_saving() {save_history = false; history_file->close();
  delete history_file;
}

void Intersection::switch_to_phase(Phase phase, bool completed_cycle) {
  // private method called from the update methods
  if(completed_cycle) {
    if(save_history) {
      (*history_file) << "Intersection (" << this->coordinates[0] << "," 
		      << this->coordinates[1] << "): " << endl;
     
      for(int i = 0; i < PHASES; i++) {
	(*history_file) << i << ": " << phase_counts[i] << " " 
			<< avg_phase_percent[i] << endl; 
      }
    }
    no_cycles++;
    cycle_duration = 0;
    reset_max_car_throughput();
    for(int i = 0; i < PHASES; i++) {
      summed_phase_durations[i] += phase_counts[i];
    }
    reset_phase_counts();
    if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) 
      cout << "cycle completed." << endl;
  }
  else {
    cycle_duration++;
    assert(cycle_duration < MAX_CYCLE_LENGTH);
  }

  if(current_phase == phase)  {
    phase_duration++;
    if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
      cout << "staying in current phase " << current_phase;
      cout << ", phase duration now " << phase_duration << endl;
    }
  }
  else {
    current_phase = phase;
    if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
      cout << "switching to phase " << current_phase << endl;
    }
    phase_duration = 0; 
  }

  for(int i = 0; i < PHASES; i++) {
    avg_phase_percent[i] = 0.99 * avg_phase_percent[i]; // decay
  }
  max_car_throughput_last_cycle[current_phase] += get_car_throughput(); 
  avg_phase_percent[current_phase] += 0.01; 
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    cout << "cycle duration now " << cycle_duration << endl;
    cout << "avg phase percent :";
    for(int i = 0; i < PHASES; i++) {
      cout << avg_phase_percent[i] << " ";
    }
    cout << endl << endl;
  }
  phase_counts[current_phase]++;
  int t = cycle_duration % RPC_CYCLE_LENGTH;
  //stat_pointer = (stat_pointer + 1) % RPC_CYCLE_LENGTH;
  summed_phase_counts[t][current_phase]++;
  //if(stat_pointer == 0) {
  //no_RPC_cycles++;
  //}
  detector_pointer = (detector_pointer + 1) % MAX_CYCLE_LENGTH;
  reset_detector_counts();
  reset_blocked();
  waiting_cars_last_step = 0;
  for(int i = 0; i < PHASES; i++) {
    waiting_times_last_step[i] = 0;
    cars_passed_last_step[i] = 0;
  }
  assert(cycle_duration < MAX_CYCLE_LENGTH);
}

void Intersection::update(Phase next_phase) {
  // Note: MixedRPController should only use FixedController, as Scats cycle completion
  // is determined differently and this method cannot know when the mixed controller
  // switches from trainer to NN.
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    cout << "updating (any next phase) ";
    print_id();
    cout <<": ";
    cout << "cycle duration is "<< cycle_duration << endl;
  }
  // (Note: cycle duration starts with 0 and goes to RPC_CYCLE_LENGTH -1)
  int remain_time = RPC_CYCLE_LENGTH -1 - cycle_duration;
  /*
  if(!(remain_time >= 0)) {
    cout <<  "cycle duration now " << cycle_duration;
    int missing_phases = 0;
    int possible_next = -1;
    for(int i = 0; i < PHASES; i++) {
      if(phase_counts[i] == 0) {
	cout << "phase "<<i<<" has not been on" << endl;
	if(possible_next == -1) {
	  possible_next = i;
	}
	missing_phases++;
      }
    }
  }
  */
  assert(remain_time >= 0);
  //if(cycle_duration >= MAX_CYCLE_LENGTH - (PHASES - 1)) {
  if(remain_time < PHASES) {
    //enforce missing phases:

    int missing_phases = 0;
    int possible_next = -1;
    for(int i = 0; i < PHASES; i++) {
      if(phase_counts[i] == 0) {
	if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
	   cout << "phase "<<i<<" has not been on" << endl;
	 }
	if(possible_next == -1) {
	  possible_next = i;
	}
	missing_phases++;
      }
    }
    
    if(missing_phases > 0) {
    //if(possible_next != -1) {
      //cout << "cycle_duration " << cycle_duration << ", remain_time "<< remain_time
      //<< ", missing phases " << missing_phases << endl;
      if(missing_phases ==  remain_time && phase_counts[next_phase] > 0) {
	if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
	  print_id();
	  cout << ": cycle duration is "<< cycle_duration <<", ";
	  cout << "suggested new phase is " << next_phase <<", ";
	  cout << "forcing to next phase " << possible_next << endl;
	  cout << "phase counts are ";
	  for(int i = 0; i < PHASES; i++) {
	    cout << phase_counts[i] << " "; 
	  }
	  cout << endl;
	}
	Phase phase = (Phase) possible_next;
	bool cycle_completed = false;
	//if(mixedRPC) {
	//cout << "&";
	//cycle_completed = (phase == starting_phase);
	//}
	return switch_to_phase(phase, cycle_completed);
      }
    }
  }
  // only reached if no forced switch
  //if(next_phase == current_phase && 
  // phase_duration == Ps::pMAX_PHASE_DURATION - 1){
  if(phase_counts[next_phase] == Ps::pMAX_PHASE_DURATION && 
     cycle_duration != RPC_CYCLE_LENGTH - 1) {
    if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    print_id();
      cout << ": max length of phase "<< next_phase <<" has been reached: " 
	   << Ps::pMAX_PHASE_DURATION  << ", cycle duration is "<< cycle_duration << endl;
    }
    next_phase = (Phase) ((next_phase + 1) % PHASES); // TODO: check if this phase 
    //isn't maxed out, too
  }  
  bool cycle_completed = cycle_duration == RPC_CYCLE_LENGTH - 1; // usual case of RPController
  //if(Ps::pCONTROLLER == STS::_MixedRPController) {
  //cout << "&";
  //cycle_completed = (next_phase == starting_phase);
  //}
  switch_to_phase(next_phase, cycle_completed);
}

void Intersection::update(bool change) {
  //TODO: implement forced switch to missing phases when MAX_CYCLE_LENGTH draws near...
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    cout << "updating ";
    print_id();
    cout <<": ";
  }
  //detector_pointer = (detector_pointer + 1) % MAX_CYCLE_LENGTH;
  //cout << "dc is "<< detector_pointer <<", ";
  //reset_detector_counts();
  if(!change && phase_duration < Ps::pMAX_PHASE_DURATION - 1) {
    switch_to_phase(current_phase, false);
  }
  else {
  if(phase_duration == (Ps::pMAX_PHASE_DURATION - 1) && 
       (Ps::pVERBOSE || _VERBOSE_INTERSECTION)) {
      cout << "max length of phase has been reached" << endl;
    }
  Phase next = (Phase) ((current_phase + 1) % PHASES);
  switch_to_phase(next, next == starting_phase);
  }
}

int Intersection::get_queue_for_turn(Turn turn) {
  switch(turn) {
  case Left:
    return LEFT_QUEUE;
  case Straight:
    return STRAIGHT_QUEUE;
  case Right:
    return RIGHT_QUEUE;
  default:
    cout << "Unrecognized turn." << endl;
    exit(1);
  }
  return 1;
}

Intersection::Phase Intersection::get_matching_phase(Direction from, Turn turn) {
  Phase required;
  if(from == North || from == South)
    required = NS_straight;
  else
    required = EW_straight;
  if(turn == Right)
    required = (Phase) (required + 1);
  return required;
}

bool Intersection::is_phase_matching(Direction from, Turn turn) {
  Phase required = get_matching_phase(from, turn); 
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    print_id();
    cout << ": current phase is " << current_phase;
    cout << ", required phase is " << required << endl;
  }
  if(current_phase == required)
    return true;
  else
    return false;
}

int Intersection::car_arrive(Car* c, Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  cars_queued[from][q]++;
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    print_id();
    cout << ": car enqueued into queue [" << from << "] [" << q << "]. ";
    cout << "New queue count is "<< cars_queued[from][q] << "." << endl;
  }
  //cars_queued_set[from][q].insert(c);
  cars_queued_list[from][q].push_back(c);
  /*
 if(coordinates[0] == 2 && coordinates[1] == 2) {
    print_id();
    cout << ": car arrived." << endl; 
  }
  */
  return cars_queued[from][q] - 1;
}

int Intersection::get_car_throughput() {
  
  if(phase_duration == 0)
    return  THROUGHPUT_START_UP;
  else
    return  THROUGHPUT_FLOW;
}

int Intersection::get_car_throughput_left(Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
 if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
   print_id();
   cout << "in car_throughput_left(),  " << from << " " << q << endl;
   cout << get_car_throughput()<< " "<<detector_counts[detector_pointer][from][q];
   cout << ", returning: ";
   cout << get_car_throughput() - detector_counts[detector_pointer][from][q] 
	<< endl;
 }
  return get_car_throughput() - detector_counts[detector_pointer][from][q];
}

bool Intersection::is_queue_full(Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  return cars_queued[from][q] >= (capacity_incoming[from] / 2);
}

bool Intersection::is_queue_empty(Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  assert(cars_queued[from][q] >= 0);
  return cars_queued[from][q] == 0;
}

void Intersection::blocked_queue(Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  blocked[from][q] = true;
}

bool Intersection::is_blocked(Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  return blocked[from][q]; 
}

bool Intersection::is_maxed_out(Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  return detector_counts[detector_pointer][from][q] >= get_car_throughput();
}

void Intersection::car_waiting() {
  waiting_cars_last_step += 1;
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    print_id();
    cout << ": car is waiting, waiting_cars is now " << waiting_cars_last_step << endl;
  }
}

bool Intersection::queue_pos_leq_throughput_left(Car* c, Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  int remain = get_car_throughput_left(from, turn);
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    if(remain == 0) {
      cout << "no throughput left " << endl;
    }
    else {
      cout << "looking at first " << remain << " cars in queue" << endl;
    }
  }
  list<Car*>::iterator pos = cars_queued_list[from][q].begin();
  for(int i = 0; i < remain; i++) {
    if(*pos == c) {
      if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
	cout << "Car is at position "<< i << " in queue" << endl;
      }
      return true;
    }
    pos++;
  }
  return false;
}

void Intersection::delete_car_from_queue_list(Car* c, Direction from, Turn turn) {
  int q = get_queue_for_turn(turn);
  int remain = get_car_throughput_left(from, turn);
  list<Car*>::iterator pos = cars_queued_list[from][q].begin();
  for(int i = 0; i < remain; i++) {
    if(*pos == c) {
      if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
	cout << "Deleting car at position "<< i << " in queue" << endl;
      }
      cars_queued_list[from][q].erase(pos);
      return;
    }
    pos++;
  }
  cout << "Car not in list!!";
  exit(1);
}

void Intersection::car_depart(Car* c, Direction from, Turn turn, double wait) {
  int q = get_queue_for_turn(turn);
  cars_queued[from][q]--;
  //cars_queued_set[from][q].erase(c);
  delete_car_from_queue_list(c, from, turn);
  advance_car(from, 0);
  detector_counts[detector_pointer][from][q]++;

  /*
  set<Car*>::iterator i;
  for(i = cars_queued_set[from][q].begin();  
      i != cars_queued_set[from][q].end(); i++) {
    (*i)->decrease_queue_number();
  }
  */
  assert(detector_counts[detector_pointer][from][q] <= 
	 THROUGHPUT_FLOW);

  waiting_time[current_phase] += wait;
  waiting_times_last_step[current_phase] += wait;
  cars_passed_last_step[current_phase] ++;
  cars_passed[current_phase]++;
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
    print_id();
    cout << ": car dequeued from queue [" << from << "] ["  << q << "], had waited  " 
	 << wait << " time steps." << endl;
    cout << "waiting_time is " << waiting_time[current_phase] << ", cars_passed is " << cars_passed[current_phase] << endl;
    cout << "New queue count is "<< cars_queued[from][q] << "." << endl;
    cout << "New detector count for last interval is "<< 
      get_detector_count_last_step(from, q)  << "." << endl;
    cout << "New detector count for last cycle is "<< 
      get_detector_count_last_cycle(from, q)  << "." << endl;
    
    cout << "New detector count is "<< 
      detector_counts[detector_pointer][from][q]  << ", detector_pointer is " <<
      detector_pointer << ", cycle duration is " << cycle_duration <<" ." << endl;
    
  }
  /*
  if(coordinates[0] == 1 && coordinates[1] == 2) {
    print_id();
    cout << ": car passed through." << endl; 
  }
  */
}
void Intersection::enter_car_on_road(Direction dir, int pos) {
  
  if(Ps::pVERBOSE)  {
    print_id();
    cout << ": adding car at " <<pos<<" in incoming dir "
	 <<dir << ", new count is " <<cars_at_segment[dir].at(pos) + 1 << endl;  
  }
  
  cars_at_segment[dir].at(pos) =  cars_at_segment[dir].at(pos) + 1;
}
void Intersection::advance_car(Direction dir, int pos) {
  
    if(Ps::pVERBOSE)  {
    print_id();
    cout << ": moving car from " <<pos<<" to "<<pos - 1<<" in incoming dir "
    <<dir << endl;  
    }
  
  cars_at_segment[dir].at(pos) =  cars_at_segment[dir].at(pos) - 1;
  if(pos > 0) {
    cars_at_segment[dir].at(pos - 1) =  cars_at_segment[dir].at(pos - 1) + 1;
  }
}

void Intersection::car_arrived_at_destination(Direction dir) {
  cars_at_segment[dir].at(0)--; 
}
 
bool  Intersection::is_road_free(Direction dir) {
  return is_road_free(dir, Ps::pROAD_LENGTH);
}

bool Intersection::is_road_free(Direction dir, int segment) {
  if(incoming[dir] == NULL) {
    cout << "Intersection::is_road_free(): No road incoming from direction "<< 
      dir << "!" << endl;
    exit(1);
  }
  
  if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) { 
    if(cars_at_segment[dir].at(segment) < capacity_incoming[dir])  {
      print_id();
      cout << ": road is free " << dir << " " << segment << " " << 
	cars_at_segment[dir].at(segment) << endl;
    }
    else {
      cout << ": road is full " << dir << " " << segment << " " << 
	cars_at_segment[dir].at(segment)<< endl;
    }
  }
  return cars_at_segment[dir].at(segment) < capacity_incoming[dir];
}

double Intersection::get_cars_passed_last_step() {
  double cars = 0;
  for(int i = 0; i < PHASES; i++) {
    cars += cars_passed_last_step[i];
  }
  return cars;
}

double Intersection::get_wait_times_last_step() {
  double wait_time = 0;
  for(int i = 0; i < PHASES; i++) {
    wait_time += waiting_times_last_step[i];
  }
  return wait_time;
}

double Intersection::get_avg_wait_time(Intersection::Phase phase){
  double r = 0;
  if(cars_passed[phase] != 0) {
    r = waiting_time[phase] / cars_passed[phase];
  }
  return r;
}

double Intersection::get_avg_wait_time(){
  double r = 0;
  double cars = 0;
  double wait_time = 0;
  for(int i = 0; i < PHASES; i++) {
    cars += cars_passed[i];
    wait_time += waiting_time[i];
  }
  if(cars_passed != 0) {
    r = wait_time / cars;
  }
  return r;
}

double Intersection::get_avg_phase_count(int time_step, Phase phase) {
  //if(coordinates[0] == 0 && coordinates[1] == 2)
  //cout << "summed_phase_counts["<<time_step<<"]["<<phase<<"]: "
  // << summed_phase_counts[time_step][phase]<<", no_cycles: "<< (no_cycles + 1)<< endl;
  assert(time_step < RPC_CYCLE_LENGTH);
  return ( (double) summed_phase_counts[time_step][phase]) / (no_cycles + 1);  // no_cycles starts from 0...
}

double Intersection::get_avg_phase_duration(Phase phase) {
  return ( (double) summed_phase_durations[phase]) / (no_cycles + 1); // no_cycles starts from 0...
}

void Intersection::reset_statistics() {
  // should be called after each reading of the statistics
  //stat_pointer = -1;
  no_cycles = 0;
  for(int i = 0; i < RPC_CYCLE_LENGTH; i++) {
    for(int j = 0; j < PHASES; j++) {
      waiting_time[j] = 0;
      cars_passed[j] = 0;
      summed_phase_durations[j] = 0;
      summed_phase_counts[i][j] = 0;
    }
  }
}

void Intersection::reset_detector_counts() {
  for(int i = 0; i < DIRECTIONS; i++)
    for(int j = 0; j < QUEUES; j++)
      detector_counts[detector_pointer][i][j] = 0;
}

void Intersection::reset_phase_counts() {
  for(int i = 0; i < PHASES; i++) {
    phase_counts[i] = 0;
  }
}

void Intersection::reset_blocked() {
for(int i = 0; i < DIRECTIONS; i++) {
    for(int j = 0; j < QUEUES; j++) {
      blocked[i][j] = false;
    }
 }
}

void Intersection::reset_max_car_throughput() {
  for(int i = 0; i < PHASES; i++)
    max_car_throughput_last_cycle[i] = 0;
}


Intersection* Intersection::get_neighbouring_intersection(Direction dir) {
  return outgoing[dir];
}

int Intersection::get_road_length(Direction dir) {
  //if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) 
  //cout << "returning new road position " <<  length_incoming[dir] << endl;
  if(incoming[dir] != NULL) {
    assert(length_incoming[dir] != -1);
    return length_incoming[dir]; 
  }
  else {
    cout << "Intersection::get_road_length() : No incoming road in direction "
	 << dir << "!" <<endl;
    exit(1);
  }
}
/*
int Intersection::get_road_capacity(Direction dir) {
  //if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) 
  //cout << "returning road capacity " <<  capacity_incoming[dir] << endl;
  if(capacity_incoming[dir] != -1) {
    return capacity_incoming[dir]; 
  }
  else {
    cout << "Intersection::get_road_capacity() : No incoming road in direction "
	 << dir << "!" <<endl;
    exit(1);
  }
}
*/

double Intersection::get_cycle_progression() {
  return ((double) cycle_duration) /  RPC_CYCLE_LENGTH;
}

int Intersection::get_detector_count(int direction, int queue, int steps) {
  int count = 0;
  assert(steps <= MAX_CYCLE_LENGTH);
  for(int i = 0; i < steps; i++) {
    int pos = detector_pointer - i;
    if(pos < 0) {
      pos += MAX_CYCLE_LENGTH;
    }
    count += detector_counts[pos][direction][queue];
  }
  return count;
}

int Intersection::get_detector_count(int direction, int queue, int from, int to) {
  int count = 0;
  //assert(intervals <= MAX_CYCLE_LENGTH);
  /*
  if(Ps::pVERBOSE) {
      print_id();
      cout << ": summing up detector count for dir " << direction << ", queue "
	   << queue << ", from "<< from<< ", to "<< to<<", pos is ";
    }
  */
  int pos  = from;
  do {
    /*
    if(Ps::pVERBOSE || _VERBOSE_INTERSECTION) {
      cout << pos << " ";
    }
    */
    //cout << "pos " << pos << " dir " << direction << " queue " << queue << endl;
    count += detector_counts[pos][direction][queue];
    pos = (pos + 1) % MAX_CYCLE_LENGTH;
  }
  while(pos!= (to + 1) % MAX_CYCLE_LENGTH);
  return count;
}

int Intersection::get_detector_count_last_step(int direction, int queue) {
  int dc =  get_detector_count(direction, queue, 1);
  assert(dc <= THROUGHPUT_FLOW);
  return dc;
}

int Intersection::get_detector_count_last_cycle(int direction, int queue) {
  if(cycle_duration >= MAX_CYCLE_LENGTH) {
    cout << "cycle duration kaputt!" << endl;
    exit(1);
  }
  return get_detector_count(direction, queue, cycle_duration+1);
}

double Intersection::get_throughput_last_step() {
  double sum = 0;
  for(int d = 0; d < DIRECTIONS; d++) {
    for(int q = 0; q < QUEUES; q++) {
      sum += get_detector_count_last_step(d, q);
    }
  }
  return sum;
}

int Intersection::get_summed_detector_count(int outgoing_dir, int from, int to) 
{
	// sums all detector counts that potentially send cars into outgoing direction
	int count = 0;
	int d = (outgoing_dir + 1) % DIRECTIONS;

	count += get_detector_count(d, RIGHT_QUEUE, from, to);
	d = (d + 1) % DIRECTIONS;
	count += get_detector_count(d, STRAIGHT_QUEUE, from, to);
	d = (d + 1) % DIRECTIONS;
	count += get_detector_count(d, LEFT_QUEUE, from, to);
	return count;
}

int Intersection::get_binary_neighbouring_detec_counts(int backlook) {
  // backlook should vary between max. road length - 1 and min. road length - 1
  int from = (detector_pointer - backlook);
  if(from < 0) {
    from += MAX_CYCLE_LENGTH;
  }
  //int to = detector_pointer;
  int to = from;
  double ew_traffic = 0;
  double ns_traffic = 0;
  if (incoming[North] != NULL) {
    ns_traffic +=  incoming[North]->get_summed_detector_count
      (South, from, to); 
  }
  if (incoming[South] != NULL) {
    ns_traffic +=  incoming[South]->get_summed_detector_count
      (North, from, to); 
  }
  if (incoming[East] != NULL) {
    ew_traffic +=  incoming[East]->get_summed_detector_count
      (West, from, to); 
  }  
  if (incoming[West] != NULL) {
    ew_traffic +=  incoming[West]->get_summed_detector_count
      (East, from, to); 
  }
/*
if(Ps::pVERBOSE) {
      if(coordinates[0] == 2 && coordinates[1] == 2 && ew_traffic > 0) {
	print_id();
	cout << ": neighb. dc in direction NS" << ns_traffic << ", ";
	cout <<  "neighb. dc in direction EW" << ew_traffic << endl;    
      }
      //}
cout << "ew_traffic: " << ew_traffic << ", ns_traffic: " << ns_traffic << endl;
*/
 return (int) (ew_traffic - ns_traffic);
}

int Intersection::get_binary_neighbouring_detec_counts(){
  
  int from = (detector_pointer - Ps::pROAD_LENGTH + 1);
  if(from < 0) {
    from += MAX_CYCLE_LENGTH;
  }
  int to = detector_pointer;
  //int to = from;
  double ew_traffic = 0;
  double ns_traffic = 0;
  if(incoming[North] != NULL) {
    ns_traffic +=  incoming[North]->get_summed_detector_count
      (South, from, to); 
  }
  if(incoming[South] != NULL) {
    ns_traffic +=  incoming[South]->get_summed_detector_count
      (North, from, to); 
  }
  if(incoming[East] != NULL) {
    ew_traffic +=  incoming[East]->get_summed_detector_count
      (West, from, to); 
  }  
  if(incoming[West] != NULL) {
    ew_traffic +=  incoming[West]->get_summed_detector_count
      (East, from, to); 
  }
/*
if(Ps::pVERBOSE) {
      if(coordinates[0] == 2 && coordinates[1] == 2 && ew_traffic > 0) {
	print_id();
	cout << ": neighb. dc in direction NS" << ns_traffic << ", ";
	cout <<  "neighb. dc in direction EW" << ew_traffic << endl;    
      }
      //}
cout << "ew_traffic: " << ew_traffic << ", ns_traffic: " << ns_traffic << endl;
*/
 return (int) (ew_traffic - ns_traffic);
}

double Intersection::get_neighbouring_detec_counts(int incoming_dir) {
  // return the sum of counts from all detectors of neighbouring intersection
  // that could send cars towards this intersection coming in in incoming direction
  // summed over window of length ROAD_LENGTH, 
  // normalized to be between 0 and 1.
  // (Note1: we still have the detector pointer from the previous time step)
  // (Note2: car arrives ROAD_LENGTH time steps after it creates detec count
  // at neighbour)
  // (Note3: we want the observation BEFORE the time step where the car 
  // arrives)
  // detector_pointer - Ps::pROAD_LENGTH + 1 is exactly before the time step 
  // the car will arrive, detec_point is the time step a detec_count is created

  int window_size = Ps::pROAD_LENGTH + 1;
  if(incoming[incoming_dir] != NULL) {
    double normalizer = window_size * (DIRECTIONS - 1) *  
      THROUGHPUT_START_UP;

    //TODO: check whether following is sensible...
    int from = (detector_pointer - Ps::pROAD_LENGTH + 1);
    if(from < 0) {
      from += MAX_CYCLE_LENGTH;
    }
    //int to = (detector_pointer + 1 - Ps::pROAD_LENGTH + 2);
    int to = detector_pointer;
    //if(to < 0) {
    //to += MAX_CYCLE_LENGTH;
    //}
    if(Ps::pVERBOSE) {
      print_id();
      cout << ": getting detector counts for incoming dir " << incoming_dir << 
	", detec pointer is "<< detector_pointer << ", from: " <<
	from << ", to " << to << endl;
    }
    double dc =  incoming[incoming_dir]->get_summed_detector_count
      ((incoming_dir + 2) % DIRECTIONS, from, to); 
    if(Ps::pVERBOSE) {
      if(coordinates[0] == 2 && coordinates[1] == 2) {
	print_id();
	cout << ": neighb. dc in direction " << incoming_dir << ": ";
	cout << dc << ",  " << dc / normalizer <<  endl;    
      }
    }
    return dc / normalizer;
  }
  else return 0;
}
//void Intersection::set_history_saving(char* file) {
//save_history = true; history_file(file);
//}
//void Intersection::unset_history_saving(save_history = false; history_file.close();)

/*
Intersection::Direction Intersection::next_direction(Direction from, 
						     Turn turn){
}
*/
