//#include "definitions.hh"

class Car;

class Intersection {

public: // public enumerations and fields (the latter for ease of creating intersections in scenario.cc)
  enum Phase {NS_straight, NS_right, EW_straight, EW_right};
  enum Direction {North, East, South, West}; 
  enum Turn {Left, Right, Straight};
  Intersection* incoming[DIRECTIONS];
  Intersection* outgoing[DIRECTIONS];
  int length_incoming[DIRECTIONS];
  int capacity_incoming[DIRECTIONS];
  bool dummy; // whether intersection is o/d node and doesn't need controller

private: // private fields
  int MAX_CYCLE_LENGTH;
  int RPC_CYCLE_LENGTH; 
  bool _VERBOSE_INTERSECTION;
  int coordinates[XY];
  Phase current_phase;
  int phase_duration; // set max
  int cycle_duration; // used for SCATS observations and statistics
  int phase_counts[PHASES]; // total duration of phases in last cycle
  Phase starting_phase;
  std::vector<int> cars_at_segment[DIRECTIONS];
  int cars_queued[DIRECTIONS][QUEUES];
  bool blocked[DIRECTIONS][QUEUES]; // set when a car in queue cannot move, unset at next time step
  //set<Car*> cars_queued_set[DIRECTIONS][QUEUES];
  list<Car*> cars_queued_list[DIRECTIONS][QUEUES];
  int (*detector_counts)[DIRECTIONS][QUEUES];
  int detector_pointer;
  int max_car_throughput_last_cycle[PHASES];  // only used for SCATS obs
  double avg_phase_percent[PHASES];
  double waiting_cars_last_step;
  double waiting_times_last_step[PHASES];
  double cars_passed_last_step[PHASES];

  // statistics: get reset when being read
  double waiting_time[PHASES];
  double cars_passed[PHASES];
  int (*summed_phase_counts)[PHASES];  // only useful if controllers don't exceed RPC_CYCLE_LENGTH
  int summed_phase_durations[PHASES];  
  int no_cycles; // cycles since last read of statistics, needed to average  
  //int stat_pointer; 


public: // public methods
  Intersection(int x_coordinate, int y_coordinate);
  ~Intersection();
  void reset();
  void delete_neighbour(Direction dir);
  void update(Phase next_phase); // only for non-SCATS learner, assumes fixed cycle length 
  void update(bool change);
  int car_arrive(Car* c, Direction from, Turn turn);
  void car_depart(Car* c, Direction from, Turn turn, double wait);
  void print_id();
  void print_connections();
  //int get_road_capacity(Direction dir);
  void enter_car_on_road(Direction dir, int pos);
  void advance_car(Direction dir, int pos);
  void car_arrived_at_destination(Direction dir);

  double get_avg_wait_time(Intersection::Phase phase); // statistics method
  double get_avg_wait_time(); // statistics method
  double get_avg_phase_count(int time_step, Phase phase); // statistics method
 double get_avg_phase_duration(Phase phase); // statistics method
  double get_cars_passed(Phase phase) {return cars_passed[phase];} // statistics method
  void reset_statistics(); // must be called after reading statistics

  double get_throughput_last_step();
  bool is_phase_matching(Direction from, Turn turn);
  bool queue_pos_leq_throughput_left(Car* c, Direction from, Turn turn);
  int get_car_throughput();
  int get_car_throughput_left(Direction from, Turn turn);
  Intersection* get_neighbouring_intersection(Direction dir);
  bool has_outgoing(Direction dir) {return outgoing[dir] != NULL;}
  int get_road_length(Direction dir);
  int get_x_coordinate() {return coordinates[X_COORDINATE];}
  int get_y_coordinate() {return coordinates[Y_COORDINATE];}
  Phase get_current_phase() {return current_phase;}
  int get_phase_duration() {return phase_duration;}
  int get_cycle_duration() {return cycle_duration;}
  double get_cycle_progression(); 
  bool is_detector_active(int direction, int queue) {
    return cars_queued[direction][queue] > 0;}
  int get_missing_phase();
  int get_detector_count_last_step(int direction, int queue);
  int get_detector_count_last_cycle(int direction, int queue);   
  int get_detector_count(int direction, int queue, int intervals);
  int get_detector_count(int direction, int queue, int from, int to);
  int get_max_throughput(Phase phase)  
  {return max_car_throughput_last_cycle[phase];}
  int get_phase_counts(int phase) {return phase_counts[phase];}
  double get_avg_phase_percent(int phase) {return avg_phase_percent[phase];}
  Phase get_neighbours_phase(int direction) {
    return incoming[direction]->get_current_phase();}
  int get_summed_detector_count(int outgoing_dir, int from, int to);
  double get_neighbouring_detec_counts(int incoming_dir);
  int get_binary_neighbouring_detec_counts();
  int get_binary_neighbouring_detec_counts(int backlook);
  bool is_road_free(Direction dir, int segment);
  bool is_road_free(Direction dir);
  bool is_maxed_out(Direction from, Turn turn); 
  void blocked_queue(Direction from, Turn turn);
  bool is_blocked(Direction from, Turn turn);
  bool is_dummy() {return dummy;}
  bool is_queue_full(Direction from, Turn turn);
  bool is_queue_empty(Direction from, Turn turn);
  void car_waiting(); // counts up waiting_time_last_step by one
  double get_waiting_cars_last_step() {return waiting_cars_last_step;}
  double get_wait_times_last_step();
  double get_cars_passed_last_step();

  int get_queue_for_turn(Turn turn);
  void set_history_saving(string filename);
  //void set_history_saving(const char* file) {save_history = true; history_file("policy");}
  void unset_history_saving();
  
private: // private methods
  void reset_detector_counts();
  void reset_phase_counts();
  void reset_max_car_throughput();
  void reset_blocked();

  Phase get_matching_phase(Direction from, Turn turn);
  void delete_car_from_queue_list(Car* c, Direction from, Turn turn);
 
  bool save_history;
  ofstream* history_file;
  void switch_to_phase(Phase phase, bool completed_cycle);
};
