#include "TreeBasics.hh"
#undef MY_DEBUG

using namespace std;

namespace tree_mcmc {


CRFApprox::CRFApprox(int *argc, char** argv) : CRFBasic(argc, argv)
{

}

	
/**
 *	Find a tree for initial sampling, result stored in init_tree
 *	Note: Construction of this tree requires discarding edges. 
 *	Only needs the adjacency matrix.
 */
void CRFApprox::build_init_tree(int root_id)
{
	int cur_tail_index, cur_head_index, cur_head_node, cur_edge;
	int *visited, *buf, i, fan_out, root;
	
	init_tree = new Tree;
	init_tree->num_node = global_num_node;
	init_tree->num_son = new int [global_num_node + 1];
	init_tree->local_to_global = new int [global_num_node + 1];
	init_tree->send = new Node * [global_num_node];
	init_tree->rec = new Node * [global_num_node];
	init_tree->order_id = new int [global_num_node];
	init_tree->tree_edge = new Edge * [global_num_node];
	init_tree->cond_edge_index = new int [global_num_node + 2];
	init_tree->cond_edge_array = NULL;
	
	visited = new int [global_num_node + 1];
	buf = new int [global_num_node + 1];
	cur_tail_index = 2;	
	cur_edge = 1;

	for(i = 1; i <= global_num_node; i++)
	{
		visited[i] = 0;
		init_tree->local_to_global[i] = i;
	}
		
	// First, randomly choose a root
	if (root_id < 0)
		root = int(ceil((rand() + 1) / ((double) RAND_MAX + 1 )* global_num_node) + 0.5);
	else
		root = root_id;
	buf[1] = root;
	visited[root] = 1;
	
	// Second from root, expand out	
	for (cur_head_index = 1; cur_head_index <= global_num_node && cur_head_index < cur_tail_index; cur_head_index ++)
	{
		cur_head_node = buf[cur_head_index];
		fan_out = 1;
		for (i = 1; i <= global_num_node; i ++)
		{
			if (adj_matrix[cur_head_node][i] && !visited[i])	 // there is an edge linking to an unvisited node
			{
				visited[i] = 1;
				init_tree->send[cur_edge] = node_array + cur_head_node;
				init_tree->rec[cur_edge] = node_array + i;

				init_tree->tree_edge[cur_edge] = edge_array + get_edge_id(cur_head_node, i);

				init_tree->order_id[cur_edge] = fan_out++;
				buf[cur_tail_index] = i;
				cur_edge ++;	cur_tail_index ++;
			}
		}

		init_tree->num_son[cur_head_node] = --fan_out;
		node_array[cur_head_node].max_num_son = fan_out;
	}
	
	// Now check whether the graph is connected. We assume it is connected.
	if(cur_head_index <= global_num_node)
	{
		printf("The input graph given by the adjacency matrix is not connected.  Exit.\n");
		exit(1);
	}

	for (cur_head_index = 1; cur_head_index <= global_num_node + 1; cur_head_index ++)
		init_tree->cond_edge_index[cur_head_index] = 1;
	
	delete []visited;
	delete []buf;
}


/**
 *	Calculate the send, rec ... for all the trees in all the partitions
 *	This is generic, once the trees and partitions are defined.
 *	Let's skip the level of build_partition
 */
void CRFApprox::build_all_trees()
{
	int i, j, k, cur_node, max_num_class;
	bool *edge_covered;
	Partition *part;
	Tree *p_tree;
	Node *p_node;

#ifdef RB_EDGE_SINGLE
	int cur_edge;
#endif

	build_init_tree();
	if (verbosity & 1)
		display_tree(init_tree, 1);

	edge_covered = new bool [global_num_edge + 1];
	for (i = 1; i <= global_num_edge; i ++)
		edge_covered[i] = false;

	for (i = 1; i <= global_num_partition; i ++)
	{
		if (verbosity & 1)
			printf("********   Partition %d    ********\n", i);

		part = partitions + i;
		for (j = 1; j <= part->num_trees; j ++)
		{
			p_tree = part->trees + j;
			build_tree(p_tree);

			for (k = 1; k < p_tree->num_node; k ++)
			{
#ifdef RB_EDGE_SINGLE
				cur_edge = p_tree->tree_edge[k] - edge_array;
				if (!edge_covered[cur_edge])
				{
					edge_covered[cur_edge] = true;
					p_tree->cal_edge_rb[k] = true;					
				}
				else
					p_tree->cal_edge_rb[k] = false;
#else
				p_tree->cal_edge_rb[k] = true;
#endif
			}
			if(verbosity & 1)
			{
				printf("********   Tree %d    of   Partition %d   ********\n", j, i);
				display_tree(part->trees + j, 1);
			}
		}
	}

	// Finally allocate the msg_buffer for all nodes after obtaining the max_num_son information
	max_num_class = 0;
	p_node = node_array + 1;

	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		if (p_node->num_class > max_num_class)
			max_num_class = p_node->num_class;
		p_node->msg_buffer = new double * [p_node->max_num_son + 1];
		for (i = 1; i <= p_node->max_num_son; i++)
			p_node->msg_buffer[i] = new double [p_node->num_class + 1];

		p_node ++;
	}
	
	sample_buf = new double [max_num_class + 1];
	rb_buf = new double * [max_num_class + 1];
	for (cur_node = 1; cur_node <= max_num_class; cur_node ++)
		rb_buf[cur_node] = new double [max_num_class + 1];

	delete [] edge_covered;
}


/**
 *	Perform inference on the whole graph (tree) using RB
 */
void CRFApprox::infer_graph()
{
	int cur_part, cur_tree, cur_node, cur_iter = 0, flag_time = 0, t_diff, loop_on;
	struct timeb t_start, t_current;
	Partition *p_part;
	Tree *p_tree;

//	int checkpoint = 9000;

//	srand(cur_step > checkpoint ? checkpoint : cur_step);

// 	if (cur_step > 1000)
// 	{
// 		verbosity = 6;
// 	}

//	cout<<rand()<<endl<<rand()<<endl<<rand()<<endl<<rand()<<endl<<rand()<<endl;

// 	if (cur_step == 7480)
// 	{
// 		verbosity = 15;
// 		write_potentials();
// 		#define MY_DEBUG
// 
// 	}
	// First, randomly sample the init_tree	
	// For the second partition, directly use the last sample of the first partition. and so on.
	for (cur_node = 1; cur_node <= init_tree->num_node; cur_node ++)
		node_array[init_tree->local_to_global[cur_node]].num_son = init_tree->num_son[cur_node];
	sample_tree(init_tree);
	
	clear_RB();

// 	if (cur_step == checkpoint)
// 	{
// 		verbosity = 6;
// 		dump_graph("spec.dat", "my_result.dat");
// 		write_potentials_grid();		
// 	}
	
	p_part = partitions + 1;
	for (cur_part = 1; cur_part <= global_num_partition; cur_part ++)
	{

#ifdef MY_DEBUG
		cout<<"After sample_tree(init_tree):"<<endl;
		display_tree(init_tree, 16);
#endif

		p_tree = p_part->trees + 1;
		
		for (cur_tree = 1; cur_tree <= p_part->num_trees; cur_tree ++)
		{
			for (cur_node = 1; cur_node <= p_tree->num_node; cur_node ++)
				node_array[p_tree->local_to_global[cur_node]].num_son = p_tree->num_son[cur_node];
			p_tree ++;
		}

		if (max_tree_time)
		{
			flag_time = 1;
			ftime(&t_start);
		}

		loop_on = 1;
		for (cur_iter = 1; cur_iter <= max_tree_iter && loop_on; cur_iter ++)
		{
			p_tree = p_part->trees + 1;
			for (cur_tree = 1; cur_tree <= p_part->num_trees; cur_tree ++)
			{
#ifdef MY_DEBUG
				cout<<"Before cal_cond_pot:"<<endl;
				display_tree(p_tree, 2);
#endif
				cal_cond_pot(p_tree);
#ifdef MY_DEBUG
				cout<<"After cal_cond_pot:"<<endl;
				display_tree(p_tree, 2);
#endif
				forward_msg_pass(p_tree);

#ifdef MY_DEBUG
				cout<<"After forward_msg_pass:"<<endl;
				display_tree(p_tree, 12);
#endif
				infer_root(p_tree);
#ifdef MY_DEBUG
				cout<<"After infer_root:"<<endl;
				display_tree(p_tree, 12);
#endif
				backward_msg_pass(p_tree);
#ifdef MY_DEBUG
				cout<<"After backward_msg_pass:"<<endl;
				display_tree(p_tree, 28);
#endif
				if (max_tree_time)
				{
					if (flag_time == 10)		// check only every 10 iterations
					{
						flag_time = 1;
						ftime(&t_current);

						t_diff = (int) ( ((t_current.time - t_start.time) << 10) + (t_current.millitm - t_start.millitm) );
						if(t_diff > max_tree_time)
						{
							loop_on = 0;
							break;
						}
					}
					else
						flag_time ++;
				}
				
				p_tree ++;
			}
		}

		p_part ++;
	}

	if(verbosity & 6)
		display_RB();

	normalize_RB();

	if(verbosity & 6)	
		display_RB();

	if (max_tree_time && verbosity & 16)
		printf("%d iterations undertaken.\n", cur_iter);

//  	if (rand() < RAND_MAX / 1000.0)
//  	{
//  		dump_graph("spec.dat", "my_result.dat");
// 		cout<<cur_step<<", Press any key to continue";
//  		cin.get();
//  	}

// 	if (cur_step == checkpoint)
// 	{		
// 		exit(1);
// 		read_potentials();
// 		cur_step++;
// 		infer_graph();		
// 	}
}


/**
 *	Perform sampling on the whole graph (tree) using RB
 */
void CRFApprox::sample_graph()
{
	int cur_part, cur_tree, cur_node, cur_iter, flag_time = 0, t_diff;
	struct timeb t_start, t_current;
	Partition *p_part;
	Tree *p_tree;

	// First, randomly sample the init_tree
	for (cur_node = 1; cur_node <= init_tree->num_node; cur_node ++)
		node_array[init_tree->local_to_global[cur_node]].num_son = init_tree->num_son[cur_node];
	sample_tree(init_tree);

	// Second, randomly pick a partition
	cur_part = int( ceil( (rand() + 1) / ((double) RAND_MAX + 1) * global_num_partition) + 0.5);
	p_part = partitions + cur_part;

	// Then stick to that partition
	p_tree = p_part->trees + 1;
	for (cur_tree = 1; cur_tree <= p_part->num_trees; cur_tree ++)
	{
		for (cur_node = 1; cur_node <= p_tree->num_node; cur_node ++)
			node_array[p_tree->local_to_global[cur_node]].num_son = p_tree->num_son[cur_node];
		p_tree ++;
	}

	if (max_tree_time)
	{
		flag_time = 1;
		ftime(&t_start);
	}

	for (cur_iter = 1; cur_iter <= max_tree_iter; cur_iter ++)
	{
		p_tree = p_part->trees + 1;
		for (cur_tree = 1; cur_tree <= p_part->num_trees; cur_tree ++)
		{
			cal_cond_pot(p_tree);
			forward_msg_pass(p_tree);
			sample_root(p_tree);
			backward_sample(p_tree);

			if (max_tree_time)
			{
				if (flag_time == 10)		// check only every 10 iterations
				{
					flag_time = 1;
					ftime(&t_current);

					t_diff = (int) ( ((t_current.time - t_start.time) << 10) + (t_current.millitm - t_start.millitm) );
					if(t_diff > max_tree_time)
						break;
				}
				else
					flag_time ++;
			}
			p_tree ++;
		}
	}
}


// A generic sampler initializer
void CRFApprox::init_tree_sampler(int *argc, char **argv)
{	
	CRFBasic::init_tree_sampler(argc, argv);

	int b_time, b_iter;

	b_iter = Utility::get_arg(&max_tree_iter, argc, argv, "-xti", 0, DFT_MAX_TREE_ITER);
	b_time = Utility::get_arg(&max_tree_time, argc, argv, "-xtt", 0, DFT_MAX_TREE_TIME);
	cout<<"max_tree_iter = "<<max_tree_iter<<", max_tree_time = "<<max_tree_time<<endl;

	if (b_iter && b_time)		// if both time and iter are specified, then use iter
	{
		printf("Warning, both max_iteration (%d) and max_time (%d) are specified.\n", max_tree_iter, max_tree_time);
		printf("Will use max_iter only.\n");
		max_tree_time = 0;
	}
	else if (!b_iter && !b_time)		// neither max_iteration nor max_time is specified, will use default time 2000 s
	{
		printf("Neither max_iteration nor max_time is specified, will use default time 2000 ms.\n");
		max_tree_time = 2000;
	}
	else if (b_iter)
	{
		printf("Will sample for %d iterations.\n", max_tree_iter);
	}
	else if (b_time)
	{
		printf("Will sample for %d ms.\n", max_tree_time);
	}	

	build_all_trees();
}


}
