//#include "TreeBasics.hh"

#include "StdBasics.hh"
#include "UBlasBasics.hh"
#include <boost/numeric/ublas/blas.hpp>

#include "CRFBasic.hh"


namespace tree_mcmc {


/**
 *	Constructor of CRFBasic.
 */
CRFBasic::CRFBasic(int *argc, char** argv)
{
	// Notice that there is NO polymorphism in constructor
	// So we delay the initialization to a separate call init_tree_sampler

	//	init_tree_sampler(argc, argv);
	buffer_debug = NULL;
	string_max_param[0] = '\0';
}


/**
 *	Initialize the data structure of a tree.
 */
void CRFBasic::build_tree(Tree *cur_tree, int root_id)
{
	int *visited, *global_buf, i, fan_out, global_root, local_num_node, local_next;
	int cur_tail_index, cur_head_index, cur_edge, global_cur_head_node;
	int *edge_assign, *count_cond_edge;
	Node *p_node;
	
	local_num_node = cur_tree->num_node;	// cur_tree->num_node is already there (filled outside)
	cur_tree->num_son = new int [local_num_node + 1];
	cur_tree->send = new Node* [local_num_node];
	cur_tree->rec = new Node* [local_num_node];
	cur_tree->order_id = new int [local_num_node];
	cur_tree->tree_edge = new Edge* [local_num_node];
	cur_tree->cal_edge_rb = new bool [local_num_node];
	
	cur_tree->cond_edge_index = new int [local_num_node + 2];	
	
	visited = new int [global_num_node + 1];
	global_buf = new int [local_num_node + 1];
	edge_assign = new int [global_num_edge + 1];
	count_cond_edge = new int [global_num_node + 1];
	
	cur_tail_index = 2;
	cur_head_index = 1;
	cur_edge = 1;
	
	for(i = 1; i <= global_num_node; i++)
	{
		visited[i] = -1;		
		count_cond_edge[i] = -1;
	}

	for(i = 1; i <= global_num_edge; i++)
		edge_assign[i] = 0;

	// First, randomly choose a root
	if (root_id < 0)
		global_root = cur_tree->local_to_global
				[int(ceil((rand() + 1) / ((double) RAND_MAX + 1) * local_num_node ) + 0.5)];
	else
		global_root = root_id;

	global_buf[1] = global_root;
	
	// Second from root, expand out	
	for(cur_head_index = 1; cur_head_index <= local_num_node && cur_head_index < cur_tail_index; cur_head_index ++)
	{
		global_cur_head_node = global_buf[cur_head_index];
		count_cond_edge[global_cur_head_node] = 0;

		fan_out = 1;
		for(local_next = 1; local_next <= local_num_node &&
								!adj_matrix[global_cur_head_node][cur_tree->local_to_global[local_next]]; local_next++) ;

		// traverse all the nodes in the whole graph
		for(i = 1; i <= global_num_node; i ++)
		{
			// if there is a link between i and global_cur_head_node
			if (adj_matrix[global_cur_head_node][i])
			{
				// if i is in the same tree
				if (local_next <= local_num_node && i == cur_tree->local_to_global[local_next])
				{
					if (visited[i] >= 0)		// already visited this node on this tree
					{
						if (adj_matrix[i][global_cur_head_node] < 0)		// traced to father node, that is ok
						{
							adj_matrix[i][global_cur_head_node] = 1;
							for(local_next++; local_next <= local_num_node &&
								!adj_matrix[global_cur_head_node][cur_tree->local_to_global[local_next]]; local_next++) ;
						}
						else			// if not father node, but still already visited, then it is not a tree (there is loop)
						{
							printf("Error, the tree is not a natural tree.  Exit.\n");
							display_tree(cur_tree, 1);
							exit(1);
						}
					}
					else
					{
						visited[i] = 1;
						adj_matrix[global_cur_head_node][i] = -1;
						cur_tree->send[cur_edge] = node_array + global_cur_head_node;
						cur_tree->rec[cur_edge] = node_array + i;
						cur_tree->tree_edge[cur_edge] = edge_array + get_edge_id(global_cur_head_node, i);

						cur_tree->order_id[cur_edge] = fan_out;
						global_buf[cur_tail_index] = i;
						cur_edge ++;			fan_out ++;			cur_tail_index++;

						// glide to the next neighbor of global_cur_head_node in current tree
						// local_to_global must be in increasing order
						for(local_next++; local_next <= local_num_node &&
								!adj_matrix[global_cur_head_node][cur_tree->local_to_global[local_next]]; local_next++) ;
					}
				} 
				else	// if i is in a different tree, then this edge is conditioned on
				{
					edge_assign[get_edge_id(global_cur_head_node, i)] = global_cur_head_node;	
					count_cond_edge[global_cur_head_node] ++;
				}
			}
		}

		// borrow the visited array to store the fan_out (number of son)
		visited[global_cur_head_node] = fan_out - 1;
	}

	// Record the cur_tree->num_son and node_arrary.max_num_son, allocate the sample_buf space
	cur_edge = -1;		// borrow the cur_edge for max_num_class
	for (i = 1; i <= local_num_node; i++) 
	{		
		global_cur_head_node = cur_tree->local_to_global[i];
		cur_tree->num_son[i] = visited[global_cur_head_node];
		p_node = node_array +global_cur_head_node;

		if(p_node->max_num_son < cur_tree->num_son[i])
			p_node->max_num_son = cur_tree->num_son[i];

		if (p_node->num_class > cur_edge)
			cur_edge = p_node->num_class;

		if (cur_tree->num_son[i] < 0)
		{
			printf("The tree is not connected.  Node %d is not visited.\n", cur_tree->local_to_global[i]);
			display_tree(cur_tree, 1);
			exit(1);
		}
	}

	// now count how many conditional edges each tree node has
	cur_edge = 1;
	for (i = 1; i <= local_num_node; i ++)
	{
		visited[cur_tree->local_to_global[i]] = i;	// now visited[] is the inverse map of local_to_global
		cur_tree->cond_edge_index[i] = cur_edge;
		cur_edge += count_cond_edge[cur_tree->local_to_global[i]];
	}

	cur_tree->cond_edge_index[local_num_node + 1] = cur_edge;

	cur_tree->cond_edge_array = new Edge * [cur_edge];

	// now fill up the cond_edge_array[] and cond_edge_orient[]
	for (i = 1; i <= global_num_edge; i++)
	{
		global_cur_head_node = edge_assign[i];	// get the tree node (global)
		if (global_cur_head_node)		// if this edge is conditioned on
		{
			// now assign edge i to local node count_cond_edge[edge_assign[i]]
			cur_head_index = cur_tree->cond_edge_index[visited[global_cur_head_node]] 
								+ (--count_cond_edge[global_cur_head_node]);
			cur_tree->cond_edge_array[cur_head_index] = edge_array + i;
		}
	}

	delete [] visited;
	delete [] global_buf;
	delete [] edge_assign;
	delete [] count_cond_edge;
}


/**
 *	Clear the RB estimations
 */
void CRFBasic::clear_RB()
{
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	double *p1, **p2;
	Node *p_node;
	Edge *p_edge;

	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		p1 = p_node->rb_est + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_node->num_class; cur_class_1 ++)		
			*(p1 ++) = 0.0;

		p_node ++;
	}

	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		p2 = p_edge->rb_est + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_edge->p_node[1]->num_class; cur_class_1 ++)
		{
			p1 = *(p2 ++) + 1;
			for (cur_class_2 = 1; cur_class_2 <= p_edge->p_node[2]->num_class; cur_class_2 ++)
				*(p1 ++) = 0.0;
		}
		p_edge ++;
	}
}


/**
 *	Normalize the RB estimations
 */
void CRFBasic::normalize_RB()
{
	// Finally, calculate the node and edge feature expectations
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	double *rb1, **rb2, sum;
	Node *p_node;
	Edge *p_edge;

	p_node = node_array + 1;
	
	// First, normalize the node RB estimates
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		sum = 0.0;
		rb1 = p_node->rb_est + 1;

		for (cur_class_1 = 1; cur_class_1 <= p_node->num_class; cur_class_1 ++)
			sum += *(rb1 ++);
		rb1 = p_node->rb_est + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_node->num_class; cur_class_1 ++)
			*(rb1 ++) /= sum;

		p_node ++;
	}

	// Second, normalize the edge RB estimates
	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{		
		sum = 0.0;
		rb2 = p_edge->rb_est + 1;

		for (cur_class_1 = 1; cur_class_1 <= p_edge->p_node[1]->num_class; cur_class_1 ++)
		{
			rb1 = *(rb2 ++) + 1;

			for (cur_class_2 = 1; cur_class_2 <= p_edge->p_node[2]->num_class; cur_class_2 ++)
				sum += *(rb1 ++);
		}

		rb2 = p_edge->rb_est + 1;

		for (cur_class_1 = 1; cur_class_1 <= p_edge->p_node[1]->num_class; cur_class_1 ++)
		{
			rb1 = *(rb2 ++) + 1;

			for (cur_class_2 = 1; cur_class_2 <= p_edge->p_node[2]->num_class; cur_class_2 ++)
				*(rb1 ++) /= sum;
		}
		
		p_edge ++;
	}
}


/**
 *	If you DO want RB, then use infer_distribution
 *	Use sample_distribution iff you do NOT need RB.
 *	Sample from exp(log_prob[]), which has num_class elements
 *	Just sample, NO calculation of the RB estimation and store in sample_buf[]
 *	It Subtracts the maximum element.
 */
int CRFBasic::sample_distribution(double *log_prob, int num_class)
{
	int cur_class;
	double *p_pot_1, *p_pot_2, sample, total;

	total = 0.0;
	p_pot_1 = log_prob + 1;
	p_pot_2 = sample_buf + 1;

	for (cur_class = 1; cur_class <= num_class; cur_class ++)
	{
		*p_pot_2 = Utility::fexp2r_check( *(p_pot_1 ++) );
		total += * (p_pot_2 ++);
	}

	// If no need of RB estimation, then one can avoid divisions
	while(1)
	{
		sample = total * (rand() + 1) / ((double) RAND_MAX + 1);
		p_pot_2 = sample_buf + 1;
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
		{
			if (sample <= *p_pot_2)
				return cur_class;
			sample -= *(p_pot_2 ++);
		}
	}

	return -1;
}


/**
 *	If you do NOT want RB, then use sample_distribution
 *	Use infer_distribution iff you need RB.
 *	Sample from exp(log_prob[]), which has num_class elements
 *	Also calculate the RB estimation and store in !!!sample_buf[]!!!
 *	It does NOT subtract the maximum element.
 */
int CRFBasic::infer_distribution(double *log_prob, int num_class)
{
	int cur_class;
	double *p_pot_1, *p_pot_2, total;

	total = 0.0;

	p_pot_1 = log_prob + 1;
	p_pot_2 = sample_buf + 1;
	for (cur_class = 1; cur_class <= num_class; cur_class ++)
	{
		*p_pot_2 = Utility::fexp2r_check( *(p_pot_1 ++) );
		total += * (p_pot_2 ++);
	}

	p_pot_2 = sample_buf + 1;
	for (cur_class = 1; cur_class <= num_class; cur_class ++)
		*(p_pot_2 ++) /= total;

	while (1)
	{
		total = (rand() + 1) / ((double) RAND_MAX + 2);

		p_pot_2 = sample_buf + 1;
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
		{
			if (total <= *p_pot_2)
				return cur_class;
			total -= *(p_pot_2 ++);
		}
	}

	return -1;
}


/**
 *	Calculate the conditional potentials of cur_tree
 */
void CRFBasic::cal_cond_pot(Tree *cur_tree)
{
	int cur_node, cur_edge, cur_class, orient;
	double *p_pot, *p_start, *p_cur;
	Node *p_node;
	Edge *p_edge;

	// loop over all nodes in the tree
	for (cur_node = 1; cur_node <= cur_tree->num_node; cur_node ++)
	{
		p_node = node_array + cur_tree->local_to_global[cur_node];		// current node

		p_cur = p_start = p_node->cond_pot + 1;
		p_pot = p_node->node_pot + 1;

		// Initialize the conditional potential buffer by the node potential
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			*(p_cur++) = *(p_pot++);

		// Add up all the contributions to conditional potential from conditioned neighbors
		for (cur_edge = cur_tree->cond_edge_index[cur_node]; 
				cur_edge < cur_tree->cond_edge_index[cur_node + 1]; cur_edge ++)			
		{
			// Current node's conditional potential buffer pointer
			p_cur = p_start;

			// Point to current node's neighbor on conditional edge
			p_edge = cur_tree->cond_edge_array[cur_edge];
			
			// Which end is the current node in of the conditional edge
			orient = 2 - (p_node == p_edge->p_node[2]);

			// Point to the row of conditional edge potentials, corresponding to neighbor's sample
			p_pot = p_edge->edge_pot[orient][p_edge->p_node[orient]->sampled_value] + 1;

			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
				*(p_cur++) += *(p_pot++);
		}
	}
}


/**
 *	Perform forward message passing (from leaves to root)
 */
void CRFBasic::forward_msg_pass(Tree *cur_tree)
{
	int cur_edge, cur_class_1, cur_class_2, cur_son, edge_orient;
	double sum, *p_pot_1, *p_pot_2, max_msg_1, max_msg_2, **pep2, *pep1;
	Edge *p_edge, **p_edge2;
	Node *p_send, *p_rec;

	p_edge2 = cur_tree->tree_edge + cur_tree->num_node - 1;
	for (cur_edge = cur_tree->num_node - 1; cur_edge >= 1; cur_edge --)
	{
		p_edge = *(p_edge2 --);

		// For forward msg passing, sender and receiver are inversed
		p_send = cur_tree->rec[cur_edge];
		p_rec = cur_tree->send[cur_edge];

		edge_orient = 2 - (p_send > p_rec);		// 1: large -> small, 2: small -> large

		// First, sender should fill up its sum-up row (summing up its messages)
		p_pot_1 = p_send->sum_msg + 1;
		p_pot_2 = p_send->cond_pot + 1;

		// Assign conditional potentials to sum_msg as initialization
		for (cur_class_1 = 1; cur_class_1 <= p_send->num_class; cur_class_1 ++)		
			*(p_pot_1 ++) = *(p_pot_2 ++);

		// Add up all the messages into sum_msg
		for (cur_son = 1; cur_son <= p_send->num_son; cur_son ++)
		{
			p_pot_1 = p_send->sum_msg + 1;
			p_pot_2 = p_send->msg_buffer[cur_son] + 1;

			for (cur_class_1 = 1; cur_class_1 <= p_send->num_class; cur_class_1 ++)
				*(p_pot_1++) += *(p_pot_2++);
		}

		// Find the maximum exponent
		max_msg_1 = NINFINITY;			// maximum exponent of double number (base 2)
		p_pot_1 = p_send->sum_msg + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_send->num_class; cur_class_1 ++)
		{
			if (*p_pot_1 > max_msg_1)
				max_msg_1 = *p_pot_1;
			p_pot_1 ++;
		}

		// Subtract the maximum exponent
		p_pot_1 = p_send->sum_msg + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_send->num_class; cur_class_1 ++)
			*(p_pot_1 ++) -= max_msg_1;

		// Receiver just puts the message into its message buffer
		// No intermediate data structure for message

		max_msg_2 = NINFINITY;
		p_pot_2 = p_rec->msg_buffer[cur_tree->order_id[cur_edge]] + 1;
		pep2 = p_edge->edge_pot[edge_orient] + 1;

		// Loop for all classes of receiver
		for (cur_class_2 = 1; cur_class_2 <= p_rec->num_class; cur_class_2 ++)
		{
			sum = 0.0;
			p_pot_1 = p_send->sum_msg + 1;
			pep1 = *pep2 + 1;

			// Compute the message
			for (cur_class_1 = 1; cur_class_1 <= p_send->num_class; cur_class_1 ++)
				sum += Utility::fexp2r_check( *(p_pot_1 ++) + *(pep1++) );

			*p_pot_2 = Utility::flog2r(sum);

			// Record the largest message element
			if (*p_pot_2 > max_msg_2)
				max_msg_2 = *p_pot_2;

			p_pot_1 ++;		pep2 ++;		p_pot_2 ++;
		}

		// Subtract the largest message element
		p_pot_2 = p_rec->msg_buffer[cur_tree->order_id[cur_edge]] + 1;
		for (cur_class_2 = 1; cur_class_2 <= p_rec->num_class; cur_class_2 ++)
			*(p_pot_2 ++) -= max_msg_2;

#ifdef MY_DEBUG

		printf("Message %d -> %d:\t", p_send - node_array, p_rec - node_array);
		p_pot_2 = p_rec->msg_buffer[cur_tree->order_id[cur_edge]] + 1;
		for (cur_class_2 = 1; cur_class_2 <= p_rec->num_class; cur_class_2 ++)
			printf("%g\t", *(p_pot_2 ++));
		printf("\n");

#endif
	}
}

/**
 *	Get a sample of the root and calculate the node marginal. 
 *	Run between forward and backward msg prop
 */
void CRFBasic::infer_root(Tree *cur_tree)
{
	int cur_class, cur_son;
	double *p_pot_1, *p_pot_2, max_pot;
	Node *p_node;

	p_node = cur_tree->send[1];

	// First, root should fill up its sum-up row (summing up its messages)

	// Copy conditional potentials to sum_msg
	p_pot_1 = p_node->sum_msg + 1;
	p_pot_2 = p_node->cond_pot + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		*(p_pot_1 ++) = *(p_pot_2 ++);

	// Add all received messages to sum_msg
	for (cur_son = 1; cur_son <= p_node->num_son; cur_son ++)
	{
		p_pot_1 = p_node->sum_msg + 1;
		p_pot_2 = p_node->msg_buffer[cur_son] + 1;

		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			*(p_pot_1++) += *(p_pot_2++);
	}

	// Subtract the largest element in sum_msg (normalization)
	max_pot = NINFINITY;
	p_pot_1 = p_node->sum_msg + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
	{
		if (*p_pot_1 > max_pot)
			max_pot = *p_pot_1;
		p_pot_1 ++;
	}

	p_pot_1 = p_node->sum_msg + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		*(p_pot_1 ++) -= max_pot;

	// Sample the root node
	// p_node->sum_msg will NOT be changed inside
	p_node->sampled_value = infer_distribution(p_node->sum_msg, p_node->num_class);

	// Fill up the RB estimation
	p_pot_1 = p_node->rb_est + 1;
	p_pot_2 = sample_buf + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		*(p_pot_1 ++) += *(p_pot_2 ++);
}


/**
 *	Get a sample of the root, run between forward and backward msg prop
 */
void CRFBasic::sample_root(Tree *cur_tree)
{
	int cur_class, cur_son;
	double *p_pot_1, *p_pot_2, max_pot;
	Node *p_node;

	p_node = cur_tree->send[1];

	// First, root should fill up its sum-up row (summing up its messages)

	// Assign conditional potentials to sum_msg
	p_pot_1 = p_node->sum_msg + 1;
	p_pot_2 = p_node->cond_pot + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		*(p_pot_1 ++) = *(p_pot_2 ++);

	// Multiply all the messages that the root has received
	for (cur_son = 1; cur_son <= p_node->num_son; cur_son ++)
	{
		p_pot_1 = p_node->sum_msg + 1;
		p_pot_2 = p_node->msg_buffer[cur_son] + 1;

		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			*(p_pot_1++) += *(p_pot_2++);
	}

	// Find the maximum exponent and subtract
	max_pot = NINFINITY;
	p_pot_1 = p_node->sum_msg + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
	{
		if (*p_pot_1 > max_pot)
			max_pot = *p_pot_1;
		p_pot_1 ++;
	}

	p_pot_1 = p_node->sum_msg + 1;
	for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)			
		*(p_pot_1 ++) -= max_pot;

#ifdef MY_DEBUG
		printf("Node %d marginal pot:\t", p_node - node_array);
		p_pot_1 = p_node->sum_msg + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			printf("%g\t", *(p_pot_1 ++));
		printf("\n");
#endif

	// Sample the root node
	// p_node->sum_msg will NOT be changed inside
	p_node->sampled_value = sample_distribution(p_node->sum_msg, p_node->num_class);
}

/**
 *	Calculate the RB estimation of edges.
 *	No need to take care of subtracting the maximum element of any ORIGINAL data structure
 *	Orientation:  1 if large -> small; 2 if small -> large
 */
void CRFBasic::fill_edge_rb(Edge *p_edge, Node *p_send, Node *p_rec, int order_id, int orient)
{
	double *p_pot_1, *p_pot_2, *p_pot_3, *p_small, *p_large, **p_pot2_1, **p_pot2_2, max_pot, sum_exp;
	int cur_class_1, cur_class_2, num_class_1, num_class_2;

	// Store sender->sum_msg[] - msg_from_receiver in sample_buf[]
	p_pot_1 = sample_buf + 1;
	p_pot_2 = p_send->sum_msg + 1;
	p_pot_3 = p_send->msg_buffer[order_id] + 1;
	for (cur_class_1 = 1; cur_class_1 <= p_send->num_class; cur_class_1 ++)
		*(p_pot_1 ++) = *(p_pot_2 ++) - *(p_pot_3 ++);

	if (orient == 1)	// if edge is from en1 -> en2
	{
		p_small = p_rec->sum_msg + 1;			// sum_msg of End_node 1 of current edge
		p_large = sample_buf + 1;				// sum_msg of End_node 2 of current edge
	}
	else				// if edge is from en2 -> en1
	{
		p_small = sample_buf + 1;				// sum_msg of End_node 1 of current edge
		p_large = p_rec->sum_msg + 1;			// sum_msg of End_node 2 of current edge
	}

	num_class_1 = p_edge->p_node[1]->num_class;
	num_class_2 = p_edge->p_node[2]->num_class;
	p_pot2_1 = p_edge->edge_pot[1] + 1;			// 1: the edge potential
	p_pot2_2 = rb_buf + 1;						// 2: RB estimation buffer
	max_pot = NINFINITY;

#ifdef MY_DEBUG
	printf("Edge RB estimation %d -> %d:\n", p_send - node_array, p_rec - node_array);
#endif

	for (cur_class_1 = 1; cur_class_1 <= num_class_1; cur_class_1 ++)
	{
		p_pot_1 = *(p_pot2_1 ++) + 1;			// 1: current edge potential
		p_pot_2 = *(p_pot2_2 ++) + 1;			// 2: current RB estimation buffer
		p_pot_3 = p_large;						// 3: sum_msg of End_node 2 of current edge

		for (cur_class_2 = 1; cur_class_2 <= num_class_2; cur_class_2 ++)
		{
			// sum_msg(end_node_1) + current_edge_pot + sum_msg(end_node_2)
			*p_pot_2 = *p_small + *(p_pot_1 ++) + *(p_pot_3 ++);

			if (*p_pot_2 > max_pot)				// record the largest edge potential
				max_pot = *p_pot_2;
#ifdef MY_DEBUG
			printf("%g\t", *p_pot_2);
#endif
			p_pot_2 ++;
		}
#ifdef MY_DEBUG
			printf("\n");
#endif
		p_small ++;
	}

	// Subtract the largest edge potential, and exponentiate
	sum_exp = 0.0;
	p_pot2_2 = rb_buf + 1;
	for (cur_class_1 = 1; cur_class_1 <= num_class_1; cur_class_1 ++)
	{
		p_pot_2 = *(p_pot2_2 ++) + 1;

		for (cur_class_2 = 1; cur_class_2 <= num_class_2; cur_class_2 ++)
		{
			*p_pot_2 = Utility::fexp2r_check(*p_pot_2 - max_pot);
			sum_exp += *(p_pot_2 ++);
		}
	}

	// Divide by the sum of RB estimator
	p_pot2_2 = rb_buf + 1;
	p_pot2_1 = p_edge->rb_est + 1;
	for (cur_class_1 = 1; cur_class_1 <= num_class_1; cur_class_1 ++)
	{			
		p_pot_1 = *(p_pot2_1 ++) + 1;
		p_pot_2 = *(p_pot2_2 ++) + 1;			

		for (cur_class_2 = 1; cur_class_2 <= num_class_2; cur_class_2 ++)
			*(p_pot_1++) += *(p_pot_2++) / sum_exp;
	}
}


/**
 *	Calculate the RB estimation of nodes.
 *	Potentials are stored in p_node->sum_msg[]
 *	!!! NOT !!! internally subtract the largest element in p_node->sum_msg[]
 *	This must be taken care of externally.
 */
void CRFBasic::fill_node_rb(Node *p_node)
{
	int cur_class, num_class;
	double sum_pot, *p_pot_1, *p_pot_2;

	// Exponentiate the log potentials
	p_pot_1 = p_node->sum_msg + 1;
	p_pot_2 = sample_buf + 1;
	sum_pot = 0.0;
	num_class = p_node->num_class;
	for (cur_class = 1; cur_class <= num_class; cur_class ++)
	{
		*p_pot_2 = Utility::fexp2r_check(*(p_pot_1 ++));
		sum_pot += *(p_pot_2 ++);
	}

	// Third, Normalize and add to RB estimation
	p_pot_1 = sample_buf + 1;
	p_pot_2 = p_node->rb_est + 1;
	for (cur_class = 1; cur_class <= num_class; cur_class ++)
		*(p_pot_2 ++) += *(p_pot_1 ++) / sum_pot;
}


/**
 *	Perform backward message passing, AND sampling (from leaves to root)
 */
void CRFBasic::backward_msg_pass(Tree *cur_tree)
{
	int cur_edge, cur_class_1, cur_class_2, edge_orient;
	double *p_pot_1, *p_pot_2, *p_pot_3, **p_pot2_1, *p_pot_3_store, max_pot, sum_pot;
	Edge **p_edge2, *p_edge;
	Node *p_send, *p_rec;

	p_edge2 = cur_tree->tree_edge + 1;
	for (cur_edge = 1; cur_edge < cur_tree->num_node; cur_edge ++)
	{
		p_edge = *(p_edge2 ++);
		p_send = cur_tree->send[cur_edge];
		p_rec = cur_tree->rec[cur_edge];
		edge_orient = 2 - (p_send > p_rec);		// 1 if large -> small; 2 if small -> large

		// 1st, sample the receiver
		p_pot_1 = *(p_edge->edge_pot[3 - edge_orient] + p_send->sampled_value) + 1;		// edge potential
		p_pot_2 = p_rec->sum_msg + 1;		// receiver's sum_msg
		p_pot_3 = sample_buf + 1;			// temp buffer
		max_pot = NINFINITY;

		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
		{
			*p_pot_3 = *(p_pot_2 ++) + *(p_pot_1 ++);
			if (*p_pot_3 > max_pot)
				max_pot = *p_pot_3;
			p_pot_3 ++;	
		}

		// Subtract the maximum exponent
		p_pot_3 = sample_buf + 1;			// temp buffer
		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
			*(p_pot_3 ++) -= max_pot;

#ifdef MY_DEBUG
		printf("Node %d Dist.:\t", p_rec - node_array);
		p_pot_3 = sample_buf + 1;			// temp buffer
		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
			printf("%g\t", *(p_pot_3 ++));
		printf("\n");
#endif

		p_rec->sampled_value = sample_distribution(sample_buf, p_rec->num_class);

		// 2nd, Calculate the edge RB estimation
		if (cur_tree->cal_edge_rb[cur_edge])
			fill_edge_rb(p_edge, p_send, p_rec, cur_tree->order_id[cur_edge], edge_orient);

		// 3rd, Calculate Sender's message and Receiver updates the sum_msg row (add the sender's message)
		max_pot = NINFINITY;
		p_pot_3 = p_rec->sum_msg + 1;						// receiver's sum message
		p_pot2_1 = p_edge->edge_pot[edge_orient] + 1;		// edge potentials

		// Get sender's message formerly received from current receiver
		p_pot_3_store = p_send->msg_buffer[cur_tree->order_id[cur_edge]] + 1;	

#ifdef MY_DEBUG
		printf("Old node %d potential: ", p_rec - node_array);
		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
			printf("%g\t", *(p_pot_3 ++));
		p_pot_3 = p_rec->sum_msg + 1;

		printf("Message %d -> %d:\t", p_send - node_array, p_rec - node_array);
#endif

		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
		{
			p_pot_1 = *(p_pot2_1 ++) + 1;			// edge potentials
			p_pot_2 = p_send->sum_msg + 1;			// sender's sum message

			// Get sender's message formerly received from current receiver
			p_pot_3_store = p_send->msg_buffer[cur_tree->order_id[cur_edge]] + 1;	

			sum_pot = 0.0;

			for (cur_class_2 = 1; cur_class_2 <= p_send->num_class; cur_class_2 ++)
				sum_pot += Utility::fexp2r_check( *(p_pot_2 ++) - *(p_pot_3_store ++) + * (p_pot_1 ++) );

			*p_pot_3 += Utility::flog2r(sum_pot);
			if (*p_pot_3 > max_pot)
				max_pot = *p_pot_3;

			p_pot_3 ++;

#ifdef MY_DEBUG
			printf("%g\t", Utility::flog2r(sum_pot));
#endif
		}

		p_pot_3 = p_rec->sum_msg + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
			*(p_pot_3 ++) -= max_pot;

		// 4th, Receiver calculates the node RB estimation
		if (cur_tree->cal_node_rb)
			fill_node_rb(p_rec);

#ifdef MY_DEBUG
		printf("New node potential: ");
		p_pot_3 = p_rec->sum_msg + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_rec->num_class; cur_class_1 ++)
			printf("%g\t", *(p_pot_3 ++));
		printf("\n");
#endif
	}
}


/**
 *	Perform backward sampling, NO backward msg passing (from leaves to root)
 *	This is much simpler than backward_msg_pass
 */
void CRFBasic::backward_sample(Tree *cur_tree)
{
	int cur_edge, cur_class;
	double *p_pot_1, *p_pot_2, *p_pot_3, max_pot;
	Edge **p_edge2, *p_edge;
	Node *p_send, *p_rec;

	p_edge2 = cur_tree->tree_edge + 1;
	for (cur_edge = 1; cur_edge < cur_tree->num_node; cur_edge ++)
	{
		p_edge = *(p_edge2 ++);
		p_send = cur_tree->send[cur_edge];
		p_rec = cur_tree->rec[cur_edge];

		p_pot_1 = *(p_edge->edge_pot[2 - (p_rec > p_send)] + p_send->sampled_value) + 1;
		p_pot_2 = p_rec->sum_msg + 1;
		p_pot_3 = sample_buf + 1;

		max_pot = NINFINITY;
		for (cur_class = 1; cur_class <= p_rec->num_class; cur_class ++)
		{
			*p_pot_3 = *(p_pot_2 ++) + *(p_pot_1 ++);
			if (*p_pot_3 > max_pot)
				max_pot = *p_pot_3;
			p_pot_3 ++;
		}

		p_pot_3 = sample_buf + 1;
		for (cur_class = 1; cur_class <= p_rec->num_class; cur_class ++)
			*(p_pot_3 ++) -= max_pot;

#ifdef MY_DEBUG
		printf("Node %d marginal pot:\t", p_rec - node_array);
		p_pot_3 = sample_buf + 1;
		for (cur_class = 1; cur_class <= p_rec->num_class; cur_class ++)
			printf("%g\t", *(p_pot_3 ++));
		printf("\n");
#endif
		p_rec->sampled_value = sample_distribution(sample_buf, p_rec->num_class);
	}
}


/**
 *	Draw a sample of tree
 *	Almost only used for init_tree
 */
void CRFBasic::sample_tree(Tree *cur_tree)
{
#ifdef MY_DEBUG
	cout<<"Before cal_cond_pot:"<<endl;
	display_tree(cur_tree, 2);
#endif
	
	cal_cond_pot(cur_tree);
#ifdef MY_DEBUG
	cout<<"After cal_cond_pot:"<<endl;
	display_tree(cur_tree, 2);
#endif

	forward_msg_pass(cur_tree);
#ifdef MY_DEBUG
	cout<<"After forward_msg_pass:"<<endl;
	display_tree(cur_tree, 12);
#endif

	sample_root(cur_tree);
#ifdef MY_DEBUG
	cout<<"After infer_root:"<<endl;
	cout<<"Root sample: "<<cur_tree->send[1]->sampled_value<<endl;
	display_tree(cur_tree, 12);
#endif

	backward_sample(cur_tree);
#ifdef MY_DEBUG
	cout<<"After backward_msg_pass:"<<endl;
	display_tree(cur_tree, 12);
#endif
}


/**
 *	Return the node array point in order to read the sampling result
 */
Node * CRFBasic::get_sample_pointer()
{
	return node_array;
}


/**
 *	allocate space for cur_observation, node_fea_para, edge_fea_para, node_fea_grad, edge_fea_grad
 */
void CRFBasic::alloc_fea()
{
	int cur_node, cur_class, cur_edge;
	double **pd2, ***pd3;
	
	// First, cur_observation
	cur_observation = new double *[global_num_node + 1];
	pd2 = cur_observation + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		*(pd2 ++) = new double [num_obs_dim + 1];

	// Second, node_fea_para[1][2][3]
	//		First index: node id;	second index: label;	third index: observation dimension
	node_fea_para = new double **[global_num_node + 1];
	pd3 = node_fea_para + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		*pd3 = new double * [num_class + 1];
		pd2 = *(pd3 ++) + 1;
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
			*(pd2 ++) = new double [num_obs_dim + 1];
	}

	// Third, node_fea_trace[1][2][3]
	//		First index: node id;	second index: label;	third index: observation dimension
	node_fea_trace = new double **[global_num_node + 1];
	pd3 = node_fea_trace + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		*pd3 = new double * [num_class + 1];
		pd2 = *(pd3 ++) + 1;
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
			*(pd2 ++) = new double [num_obs_dim + 1];
	}

	// Fourth, node_fea_exp[1][2][3]
	//		First index: node id;	second index: label;	third index: observation dimension
	node_fea_exp = new double **[global_num_node + 1];
	pd3 = node_fea_exp + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		*pd3 = new double * [num_class + 1];
		pd2 = *(pd3 ++) + 1;
		for (cur_class = 1; cur_class <= num_class; cur_class ++)
			*(pd2 ++) = new double [num_obs_dim + 1];
	}

	if (num_node_extra_fea)
	{
		// Second_2, node_extra_fea_para[1][2]
		//	  First index: node id;		second index: feature index
		node_extra_fea_para = new double * [global_num_node + 1];
		pd2 = node_extra_fea_para + 1;
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
			*(pd2 ++) = new double [num_node_extra_fea + 1];

		// Third_2, node_extra_fea_trace[1][2]
		//	  First index: node id;		second index: feature index
		node_extra_fea_trace = new double * [global_num_node + 1];
		pd2 = node_extra_fea_trace + 1;
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
			*(pd2 ++) = new double [num_node_extra_fea + 1];

		// Fourth_2, node_extra_fea_exp[1][2]
		//	  First index: node id;		second index: feature index
		node_extra_fea_exp = new double * [global_num_node + 1];
		pd2 = node_extra_fea_exp + 1;
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
			*(pd2 ++) = new double [num_node_extra_fea + 1];
	}

	// Fifth, edge_fea_para[1][2]
	//	  First index: edge id;		second index: feature index
	edge_fea_para = new double * [global_num_edge + 1];
	pd2 = edge_fea_para + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
		*(pd2 ++) = new double [num_edge_fea + 1];

	// Sixth, edge_fea_trace[1][2]
	//	  First index: edge id;		second index: feature index
	edge_fea_trace = new double * [global_num_edge + 1];
	pd2 = edge_fea_trace + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
		*(pd2 ++) = new double [num_edge_fea + 1];

	// Seventh, edge_fea_exp[1][2]
	//	  First index: edge id;		second index: feature index
	edge_fea_exp = new double * [global_num_edge + 1];
	pd2 = edge_fea_exp + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
		*(pd2 ++) = new double [num_edge_fea + 1];
}


/**
 *	Allocate the space of edges and nodes for running MCMC
 */
void CRFBasic::alloc_potentials()
{
	int cur_item, cur_class_1, cur_class_2, cur_order;
	Node *p_node;
	Edge *p_edge;

	node_array = new Node[global_num_node + 1];
	edge_array = new Edge[global_num_edge + 1];

	p_node = node_array + 1;
	for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
	{
		p_node->cond_pot = new double [num_class + 1];
		p_node->rb_est = new double [num_class + 1];
		p_node->node_pot = new double [num_class + 1];
		p_node->sum_msg = new double [num_class + 1];
		p_node->num_class = num_class;
		p_node->max_num_son = 0;
		p_node ++;
	}

	p_edge = edge_array + 1;
	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
	{
		p_edge->rb_est = new double * [num_class + 1];
		for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
			p_edge->rb_est[cur_class_1] = new double [num_class + 1];

		for (cur_order = 1; cur_order <= 2; cur_order ++)
		{
			p_edge->edge_pot[cur_order] = new double * [num_class + 1];
			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
			{
				p_edge->edge_pot[cur_order][cur_class_1] = new double [num_class + 1];
				for (cur_class_2 = 1; cur_class_2 <= num_class; cur_class_2 ++)
					p_edge->edge_pot[cur_order][cur_class_1][cur_class_2] = 0.0;
			}
		}
		p_edge ++;
	}
}


/**
 *	Randomly initialize the edge_fea_para[1][2]
 *			First index: edge id;	second index: feature index
 *	For debug only.
 */
void CRFBasic::set_cur_obs_debug()
{
	int cur_node, cur_obs_dim;
	double **p_obs2, *p_obs1;
	
	p_obs2 = cur_observation + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		p_obs1 = *(p_obs2 ++) + 1;

		for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			*(p_obs1 ++) = 0.0;		// cur_observation[cur_node][cur_obs_dim] = 0;
	}

	cur_observation[1][2] = 1;
}


/**
 *	Fill up the node_fea_para[1][2][3]
 *			First index: node id;	second index: label;	third index: observation 
 *	For debug only.
 */
void CRFBasic::set_node_fea_para_debug()
{
	int cur_node, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1;
	Node *p_node;
	
	nfp3 = node_fea_para + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				*(nfp1 ++) = (cur_class == cur_obs_dim);
			}
		}
		p_node ++;
	}
}


/**
 *	Fill up the edge_fea_para[1][2]
 *			First index: edge id;	second index: feature index
 *	For debug only.
 */
void CRFBasic::set_edge_fea_para_debug()
{
	int cur_edge, cur_fea;
	double **efp2, *efp1;

	efp2 = edge_fea_para + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
		{
			*(efp1 ++) = -1;		// Encourage edge ends to be different
		}
	}
}


/**
 *	Randomly initialize the node_fea_para[1][2][3]
 *			First index: node id;	second index: label;	third index: observation
 */
void CRFBasic::randomize_node_fea_para(double range)
{
	int cur_node, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1;
	Node *p_node;
	
	if (range >= 0)
		maxNodeRand = range;

	nfp3 = node_fea_para + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				*(nfp1 ++) = maxNodeRand / RAND_MAX * (rand() - RAND_MAX / 2) * 2;
//				*(nfp1 ++) = 1;
			}
		}
		p_node ++;
	}

//	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
//		node_fea_para[cur_node][3][1] = 30;
// changed
//	node_fea_para[2][3][1] = 100;

	//////////////////////////////////////////////////////////////////////////
	// For node extra feature
	if (num_node_extra_fea)
	{
		nfp2 = node_extra_fea_para + 1;		
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_class = 1; cur_class <= num_node_extra_fea; cur_class ++)
				*(nfp1 ++) = maxNodeRand / RAND_MAX * (rand() - RAND_MAX / 2) * 2;
		}
	}
}


/**
 *	Randomly initialize the edge_fea_para[1][2]
 *			First index: edge id;	second index: feature index
 */
void CRFBasic::randomize_edge_fea_para(double range)
{
	int cur_edge, cur_fea;
	double **efp2, *efp1;

	efp2 = edge_fea_para + 1;
	if (range >= 0)
		maxEdgeRand = range;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)		
			*(efp1 ++) = maxEdgeRand / RAND_MAX * (rand() - RAND_MAX / 2) * 2;
//			*(efp1 ++) = 1;
/*
		if (cur_edge > 1 && cur_edge < global_num_edge)
		{
			efp1 = *(efp2 - 1); 
			for (cur_fea = 1; cur_fea <= num_class; cur_fea ++)
			{
				*(efp1 + (cur_fea - 1) * num_class + cur_fea) = 100;
			}
		}
*/
	}
}


/**
 *	Reset the trace elements to 0
 */
void CRFBasic::reset_trace() 
{ 
	int cur_node, cur_fea, cur_edge, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;
	
	//  First, nodes
	nfp3 = node_fea_trace + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				*(nfp1 ++) = 0;
			}
		}
		p_node ++;
	}

	// If possible, extra node features
	if (num_node_extra_fea)
	{
		nfp2 = node_extra_fea_trace + 1;		
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
				*(nfp1 ++) = 0;
		}
	}

	//  Second, edges
	efp2 = edge_fea_trace + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
		{
			*(efp1 ++) = 0;
		}
	}
}


/**
 *	Discount the trace by factor discount
 */
void CRFBasic::discount_trace(double discount) 
{
	int cur_node, cur_fea, cur_edge, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;

	//  First, nodes
	nfp3 = node_fea_trace + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				*(nfp1 ++) *= discount;		 // update node feature grads, node_fea_para[i][j][k]
			}
		}
		p_node ++;
	}

	// If possible, extra node features
	if (num_node_extra_fea)
	{
		nfp2 = node_extra_fea_trace + 1;		
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
				*(nfp1 ++) *= discount;
		}
	}

	//  Second, edges
	efp2 = edge_fea_trace + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
		{
			*(efp1 ++) *= discount;			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
		}
	}
}


/**
 *	Re-initialize randomly the feature parameters
 */
void CRFBasic::reset_params() 
{
	randomize_node_fea_para();
	randomize_edge_fea_para();	
}


/**
 *	Calculate the node feature expectations.
 */
void CRFBasic::cal_node_fea_exp() 
{
	// Calculate the node feature expectations
	int cur_node, cur_fea, cur_class_1;
	double *rb1, **cob2, ***fept3, **fept2, *fept1, *cob1;
	Node *p_node;

	p_node = node_array + 1;
	fept3 = node_fea_exp + 1;

	if (step_size[1] > 1e-100)
	{
		cob2 = cur_observation + 1;
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			fept2 = *(fept3 ++) + 1;
			rb1 = p_node->rb_est + 1;

			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
			{
				fept1 = *(fept2 ++) + 1;

				cob1 = (*cob2) + 1;
				for (cur_fea = 1; cur_fea <= num_obs_dim; cur_fea ++)
					*(fept1 ++) = *(cob1 ++) * (*rb1);

				rb1 ++;
			}

			p_node ++;		cob2 ++;
		}
	}
	else
	{
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			fept2 = *(fept3 ++) + 1;

			for (cur_class_1 = 1; cur_class_1 <= p_node->num_class; cur_class_1 ++)
			{
				fept1 = *(fept2 ++) + 1;

				for (cur_fea = 1; cur_fea <= num_obs_dim; cur_fea ++)
					*(fept1 ++) = 0;
			}
		}	
	}
}


/**
 *	Calculate the node feature traces.
 */
void CRFBasic::fill_node_trace() 
{
	int cur_item, cur_class, cur_obs_dim;
	double ***pd3, **pd2, *pd1, ***pd3_2, **pd2_2, *pd1_2, **pd2_3, *pd1_3;
	Node *p_node;

	//  First, nodes
	if (step_size[1] > 1e-100)
	{
		pd3 = node_fea_trace + 1;
		p_node = node_array + 1;
		pd3_2 = node_fea_exp + 1;
		pd2_3 = cur_observation + 1;

		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			pd2 = *(pd3 ++) + 1;
			pd2_2 = *(pd3_2 ++) + 1;

			for (cur_class = 1; cur_class < p_node->sampled_value; cur_class ++)
			{
				pd1 = *(pd2 ++) + 1;
				pd1_2 = *(pd2_2 ++) + 1;

				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
				{
					*pd1_2 = - *pd1_2;
					*(pd1 ++) += *(pd1_2 ++);
				}
			}

			pd1 = *(pd2 ++) + 1;
			pd1_2 = *(pd2_2 ++) + 1;
			pd1_3 = *(pd2_3 ++) + 1;

			// cur_class == p_node->sampled_value
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)			
			{
				*pd1_2 = *(pd1_3 ++) - *pd1_2;
				*(pd1 ++) += *(pd1_2 ++);
			}

			for (cur_class ++; cur_class <= p_node->num_class; cur_class ++)
			{
				pd1 = *(pd2 ++) + 1;
				pd1_2 = *(pd2_2 ++) + 1;

				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
				{
					*pd1_2 = - *pd1_2;
					*(pd1 ++) += *(pd1_2 ++);
				}
			}
			p_node ++;
		}
	}
}


// Set the edge and node potentials of the grid.  
// Space allocation has been done by alloc_potentials_grid()
void CRFBasic::set_potentials_debug()
{
	int cur_node, cur_edge, cur_fea, cur_class_1;
	Node *p_node_1;
	Edge *p_edge;
	double *cur_pot1, sum, **obs2, *obs1, ***nfp3, **nfp2, *nfp1, **efp2;

/*	For most general way to define edge potentials
	double **cur_pot2_1, **cur_pot2_2, *efp1;
	Node *p_node_2;
	int cur_class_2;
*/

	// First Node potentials
	p_node_1 = node_array + 1;
	obs2 = cur_observation + 1;
	nfp3 = node_fea_para + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		cur_pot1 = p_node_1->node_pot + 1;
		nfp2 = (*nfp3) + 1;

		for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
		{
			sum = 0.0;
			obs1 = (*obs2) + 1;
			nfp1 = (*nfp2) + 1;
			for (cur_fea = 1; cur_fea <= num_obs_dim; cur_fea ++)
			{
				//	sum += cur_observation[cur_node][cur_fea] * node_fea_para[cur_node][cur_class][cur_fea];
				sum += (*obs1) * (*nfp1);
				obs1 ++;		nfp1++;
			}

			*cur_pot1 = sum;
			cur_pot1 ++;		nfp2 ++;
		}

		p_node_1 ++;
		obs2 ++;	nfp3 ++;
	}

	// Second Edge potentials
	p_edge = edge_array + 1;
	efp2 = edge_fea_para + 1;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)		// loop over edges
	{
		for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
		{
			p_edge->edge_pot[1][cur_class_1][cur_class_1] = (*efp2)[1];
			p_edge->edge_pot[2][cur_class_1][cur_class_1] = (*efp2)[1];
		}

		p_edge ++;	efp2 ++;
	}	
}


void CRFBasic::self_test()
{
	set_cur_obs_debug();
	set_node_fea_para_debug();
	set_edge_fea_para_debug();

	for(int i = 0; i < 1000; i ++)
	{
		set_potentials_debug();
		cur_step ++;
		infer_graph();
	}

	set_potentials_debug();
	cur_step ++;
	infer_graph();
	exit(1);
}


void CRFBasic::set_node_potentials()
{
	int cur_item, cur_fea, cur_class_1, cur_class_2;
	Node *p_node_1;
	double *cur_pot1, sum, **obs2, *obs1, ***nfp3, **nfp2, *nfp1;
	Edge *p_edge;

	//////////////////////////////////////////////////////////////////////////
	// First Node potentials

	p_node_1 = node_array + 1;
	obs2 = cur_observation + 1;
	nfp3 = node_fea_para + 1;
	if (step_size[1] > 1e-100)
	{
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			cur_pot1 = p_node_1->node_pot + 1;
			nfp2 = (*nfp3) + 1;

			for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
			{
				sum = 0.0;
				obs1 = (*obs2) + 1;
				nfp1 = (*nfp2) + 1;
				for (cur_fea = 1; cur_fea <= num_obs_dim; cur_fea ++)
				{
					//	sum += cur_observation[cur_node][cur_fea] * node_fea_para[cur_node][cur_class][cur_fea];
					sum += (*obs1) * (*nfp1);
					obs1 ++;		nfp1++;
				}

				if (maxNodeRand < 0)
					cout<<cur_class_1<<"\t"<<sum<<endl;
				*(cur_pot1 ++) = sum;
				nfp2 ++;
			}
			
			if (fabs(bias) > 1e-5)
				p_node_1->node_pot[cur_step % cycleTime / stepsPerControl + 1] += bias;

			p_node_1 ++;
			obs2 ++;	nfp3 ++;
		}
	}
	else
	{
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			cur_pot1 = p_node_1->node_pot + 1;

			for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
				*(cur_pot1 ++) = 0.0;
			
			if (fabs(bias) > 1e-5)
				p_node_1->node_pot[cur_step % cycleTime / stepsPerControl + 1] += bias;

			p_node_1 ++;
		}
	}

	//////////////////////////////////////////////////////////////////////////
	//  Clear edge potentials

	p_edge = edge_array + 1;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)		// loop over edges
	{
		for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
			for (cur_class_2 = 1; cur_class_2 <= num_class; cur_class_2 ++)
			{
				p_edge->edge_pot[1][cur_class_1][cur_class_2] 
					= p_edge->edge_pot[2][cur_class_1][cur_class_2] 
					= 0;
			}

		p_edge ++;
	}
}


void CRFBasic::update_fea_para(double *reward, int source)
{
	int cur_item, cur_fea, cur_class, cur_obs_dim;
	double ***pd3, **pd2, *pd1, ***pd3_2, **pd2_2 = NULL, *pd1_2, cur_mult, *cur_reward;
	Node *p_node;
	
	if (source == 1)		// from trace
		pd3_2 = node_fea_trace + 1;
	else if (source == 2)	// from grad (by NAC)
		pd3_2 = node_fea_exp + 1;
	else
	{
		cout<<"Unknown source of update_fea_para: "<<source<<".  Can be either 1 for trace, or 2 for grad."<<endl;
		exit(1);
	}

	if (use_local_reward)		// Use local rewards
	{
		if (source == 1)
			fill_edge_reward(reward);
		
		//  First, nodes
		if (step_size[1] > 1e-100)
		{
			pd3 = node_fea_para + 1;
			p_node = node_array + 1;
			cur_reward = reward + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				cur_mult = *(cur_reward ++) * step_size[1];
				pd2 = *(pd3 ++) + 1;
				pd2_2 = *(pd3_2 ++) + 1;

				for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
				{
					pd1 = *(pd2 ++) + 1;
					pd1_2 = *(pd2_2 ++) + 1;

					for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
						*(pd1 ++) += *(pd1_2 ++) * cur_mult;
				}
				p_node ++;
			}
		}
		else
			cur_reward = reward + global_num_node + 1;

		//////////////////////////////////////////////////////////////////////////
		// For extra node feature
		if (num_node_extra_fea)
		{
			if (source == 1)			// from trace
				pd2_2 = node_extra_fea_trace + 1;
			else if (source == 2)		// from grad (by NAC)
				pd2_2 = node_extra_fea_exp + 1;

			pd2 = node_extra_fea_para + 1;
			cur_reward = reward + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				cur_mult = *(cur_reward ++) * step_size[3];
				pd1 = *(pd2 ++) + 1;
				pd1_2 = *(pd2_2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					*(pd1 ++) += *(pd1_2 ++) * cur_mult;
			}
		}

		//  Second, edges
		if (source == 1)		// from trace
			pd2_2 = edge_fea_trace + 1;
		else if (source == 2)	// from grad (by NAC)
			pd2_2 = edge_fea_exp + 1;

		pd2 = edge_fea_para + 1;
	
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			cur_mult = *(cur_reward ++) * step_size[2];
			pd1 = *(pd2 ++) + 1;
			pd1_2 = *(pd2_2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea++)
				*(pd1 ++) += *(pd1_2 ++) * cur_mult;
		}	
	}
	else		// Use global rewards
	{
		//  First, nodes
		cur_reward = reward + 1;
		if (step_size[1] > 1e-100)
		{
			pd3 = node_fea_para + 1;
			p_node = node_array + 1;
			cur_mult = step_size[1] * (*cur_reward);
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				pd2 = *(pd3 ++) + 1;
				pd2_2 = *(pd3_2 ++) + 1;

				for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
				{
					pd1 = *(pd2 ++) + 1;
					pd1_2 = *(pd2_2 ++) + 1;

					for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
						*(pd1 ++) += *(pd1_2 ++) * cur_mult;
				}
				p_node ++;
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// For extra node feature
		if (num_node_extra_fea)
		{
			cur_mult = step_size[3] * (*cur_reward);
			if (source == 1)			// from trace
				pd2_2 = node_extra_fea_trace + 1;
			else if (source == 2)		// from grad (by NAC)
				pd2_2 = node_extra_fea_exp + 1;

			pd2 = node_extra_fea_para + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				pd1 = *(pd2 ++) + 1;
				pd1_2 = *(pd2_2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					*(pd1 ++) += *(pd1_2 ++) * cur_mult;
			}
		}

		//  Second, edges
		if (source == 1)		// from trace
			pd2_2 = edge_fea_trace + 1;
		else if (source == 2)	// from grad (by NAC)
			pd2_2 = edge_fea_exp + 1;

		cur_mult = step_size[2] * (*cur_reward);
		pd2 = edge_fea_para + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			pd1 = *(pd2 ++) + 1;
			pd1_2 = *(pd2_2 ++) + 1;

			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				*(pd1 ++) += *(pd1_2 ++) * cur_mult;
		}	
	}
}


void CRFBasic::graph_para_write(std::ostream & o) 
{
	int cur_node, cur_fea, cur_class, cur_edge, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;

	//  First, nodes
	nfp3 = node_fea_para + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				o << *(nfp1 ++) << "  ";	 // update node feature grads, node_fea_para[i][j][k]
			}
			o << endl;
		}
		o << endl;

		p_node ++;
	}

	// If possible, extra node features
	if (num_node_extra_fea)
	{
		nfp2 = node_extra_fea_para + 1;		
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_class = 1; cur_class <= num_node_extra_fea; cur_class ++)
				o << *(nfp1 ++) << "  ";
			o << endl;
		}
	}

	//  Second, edges
	efp2 = edge_fea_para + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
		{
			o << *(efp1 ++) << "  ";			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
		}
		o << endl << endl;
	}
}


void CRFBasic::graph_para_read(std::istream & o) 
{
	int cur_node, cur_fea, cur_edge, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;

	//  First, nodes
	nfp3 = node_fea_para + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				o >> *(nfp1 ++);	 // update node feature grads, node_fea_para[i][j][k]
			}
		}
		p_node ++;
	}

	// If possible, extra node features
	if (num_node_extra_fea)
	{
		nfp2 = node_extra_fea_para + 1;		
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_class = 1; cur_class <= num_node_extra_fea; cur_class ++)
				o >> *(nfp1 ++);
		}
	}

	//  Second, edges
	efp2 = edge_fea_para + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
		{
			o >> *(efp1 ++);			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
		}
	}
}


/**
 *	Get the maximum feature parameter
 */
double CRFBasic::get_max_param() 
{
	int cur_node, cur_fea, cur_edge, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1, record, record1;
	Node *p_node;

	record = record1 = -1;
	string_max_param[0] = '\0';

	//  First, nodes
	nfp3 = node_fea_para + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		nfp2 = *(nfp3 ++) + 1;
		for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
			{
				// node_fea_para[cur_node][cur_class][cur_obs_dim];
				if(fabs(*nfp1) > record1)
					record1 = fabs(*nfp1);
				nfp1 ++;			
			}
		}
		p_node ++;
	}

	sprintf(string_max_param, "%5.4g", record1);
	if (record1 > record)
		record = record1;
	record1 = -1;

	//  Second, edges
	efp2 = edge_fea_para + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		efp1 = *(efp2 ++) + 1;
		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
		{
			if(fabs(*efp1) > record1)
				record1 = fabs(*efp1);
			efp1 ++;
		}
	}

	sprintf(string_max_param, "%s, %5.4g", string_max_param, record1);
	if (record1 > record)
		record = record1;

	// If possible, extra node features
	if (num_node_extra_fea)
	{
		record1 = -1;
		nfp2 = node_extra_fea_para + 1;
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			nfp1 = *(nfp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
			{
				if(fabs(*nfp1) > record1)
					record1 = fabs(*nfp1);
				nfp1 ++;			
			}		
		}
		sprintf(string_max_param, "%s, %5.4g", string_max_param, record1);
	}
	if (record1 > record)
		record = record1;

	return record;
}


/**
 *	Set the step size of CRF
 */
void CRFBasic::set_step_size(double *new_step_size)
{	
	for (int i = 1; i <= 3; i ++)
		step_size[i] = new_step_size[i];
}


/**
 *	A generic sampler initializer
 */
void CRFBasic::init_tree_sampler(int *argc, char **argv)
{	
	Utility::get_arg(&num_obs_dim, argc, argv, "-no", 1, 0);
	Utility::get_arg(&num_class, argc, argv, "-ncl", 1, 0);
	Utility::get_arg(&num_edge_fea, argc, argv, "-nef", 1, 0);
	Utility::get_arg(&num_node_extra_fea, argc, argv, "-nnxf", 0, DFT_NUM_NODE_EXTRA_FEA);

	Utility::get_arg(&maxNodeRand, argc, argv, "-xn", 0, DFT_MAXNODERAND);
	Utility::get_arg(&maxEdgeRand, argc, argv, "-xe", 0, DFT_MAXEDGERAND);
	Utility::get_arg(&edge_fea_intensity, argc, argv, "-et", 0, DFT_EDGE_FEA_INTENSITY);

	cout<<"num_obs_dim = "<<num_obs_dim<<"\n"
		<<"num_class = "<<num_class<<"\n"
		<<"num_edge_fea = "<<num_edge_fea<<"\n"
		<<"num_node_extra_fea = "<<num_node_extra_fea<<"\n"
		<<"maxNodeRand = "<<maxNodeRand<<"\n"
		<<"maxEdgeRand = "<<maxEdgeRand<<"\n"
		<<"edge_fea_intensity = "<<edge_fea_intensity<<"\n";

	init_graph(argc, argv);
	alloc_fea();
	randomize_node_fea_para();
	randomize_edge_fea_para();

	cout<<"global_num_edge = "<<global_num_edge<<endl;
	Utility::get_arg(&verbosity, argc, argv, "-v", 0, DFT_VERBOSITY);

	Utility::get_arg(&use_local_reward, argc, argv, "-ulr", 0, DFT_USE_LOCAL_REWARD);	
	if (use_local_reward)
	{
		Utility::get_arg(&edge_reward_def, argc, argv, "-erd", 0, DFT_EDGE_REWARD_DEF);
		cout<<"use_local_reward = true, edge_reward_def = "<<edge_reward_def<<endl;
		rewards = new double [global_num_edge + global_num_node + 1];
	}
	else
	{
		rewards = new double [2];
		cout<<"use_local_reward = false"<<endl;
	}

	cur_step = 0;
	last_action = new int [global_num_node + 1];
}


/************************************************************************/
/*            Display functions for debugging                           */
/************************************************************************/

/**
 *	Dump the graph parameters
 */
void CRFBasic::dump_graph_specification(char * spec_file)
{
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	FILE * p_file_spec;
	Edge *p_edge;

	// First output the graph parameters (topology and potentials)
	if(! spec_file)
		p_file_spec = fopen("graph_spec.dat", "w");
	else
		p_file_spec = fopen(spec_file, "w");

	fprintf(p_file_spec, "<method> (exact_bp)\n");
	fprintf(p_file_spec, "<number of vertices> ( %d )\n", global_num_node);
	fprintf(p_file_spec, "<model type> (pairwise-discrete)\n");
	fprintf(p_file_spec, "<random variable sizes> ");
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		fprintf(p_file_spec, "( %d %d ) ", cur_node, node_array[cur_node].num_class);
	fprintf(p_file_spec, "\n");

	// Write graph edges
	fprintf(p_file_spec, "<link> ");
	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		fprintf(p_file_spec, "( %d %d ) ", p_edge->p_node[1] - node_array, p_edge->p_node[2] - node_array);
		p_edge ++;
	}

	// Write edge potentials
	fprintf(p_file_spec, "<potential tables> ");
	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		fprintf(p_file_spec, "( %d %d ) ", p_edge->p_node[1] - node_array, p_edge->p_node[2] - node_array);
		fprintf(p_file_spec, "( ");
		for (cur_class_1 = 1; cur_class_1 <= p_edge->p_node[1]->num_class; cur_class_1 ++)
		{
			for (cur_class_2 = 1; cur_class_2 <= p_edge->p_node[2]->num_class; cur_class_2 ++)			
				fprintf(p_file_spec, "%g ", pow(2, p_edge->edge_pot[2][cur_class_1][cur_class_2]));	// [2] due to a bug in PGM
		}
		fprintf(p_file_spec, ") ");
		p_edge ++;
	}
	fprintf(p_file_spec, "\n");

	fprintf(p_file_spec, "<observations>\n");
	fprintf(p_file_spec, "<internal potentials> ");

	// Write node potentials
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		fprintf(p_file_spec, "( %d ) ( ", cur_node);
		for (cur_class_1 = 1; cur_class_1 <= node_array[cur_node].num_class; cur_class_1 ++)
			fprintf(p_file_spec, "%g ", pow(2, node_array[cur_node].node_pot[cur_class_1]));
		fprintf(p_file_spec, ") ");
	}
	fprintf(p_file_spec, "\n");

	fclose(p_file_spec);

	

// now write potentials
	write_potentials();



}


// Write the potentials to file or buffer_debug, for debug purpose
// target = 1: write to file
// target = 2: write to buffer_debug
void CRFBasic::write_potentials(int target)
{
	int cur_node, cur_edge, cur_class_1;
	Node *p_node_1;
	Edge *p_edge;
	double *cur_pot1;
	FILE *fp_in;
	int cur_index = 1;

	if (target == 1)
	{
		fp_in = fopen("rec_pot.dat", "w");
		if (!fp_in)
		{
			cout<<"Cannot open rec_pot.dat"<<endl;
			exit(1);
		}
	}
	else if (target == 2)
	{
		if (!buffer_debug)
			buffer_debug = new double [3000];
	}

	// First Node potentials
	p_node_1 = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		cur_pot1 = p_node_1->node_pot + 1;

		if (target == 1)
		{
			for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
				fprintf(fp_in, "%.20lg\n", *(cur_pot1 ++));
		}
		else if (target == 2)
		{
			for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
				buffer_debug[cur_index++] = *(cur_pot1 ++);
		}

		p_node_1 ++;
	}

	// Second Edge potentials
	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)		// loop over edges
	{
		if (target == 1)
		{
			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				fprintf(fp_in, "%.20lg\n", p_edge->edge_pot[1][cur_class_1][cur_class_1]);
		}
		else if (target == 2)
		{
			for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
				buffer_debug[cur_index++] = p_edge->edge_pot[1][cur_class_1][cur_class_1];

		}

		p_edge ++;
	}	

	if (target == 1)
		fclose(fp_in);
}


// Read the potentials from file or buffer_debug, for debug purpose
// target = 1: read from file
// target = 2: read from buffer_debug
void CRFBasic::read_potentials(int target)
{
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	Node *p_node_1;
	Edge *p_edge;
	double *cur_pot1;
	FILE *fp_in;
	int cur_index = 1;

	if (target == 1)
	{
		fp_in = fopen("rec_pot.dat", "r");
		if (!fp_in)
		{
			cout<<"Cannot open rec_pot.dat"<<endl;
			exit(1);
		}
	}

	// First Node potentials
	p_node_1 = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		cur_pot1 = p_node_1->node_pot + 1;

		for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
		{
			if (target == 1)
				fscanf(fp_in, "%lg\n", cur_pot1);
			else if (target == 2)
				*cur_pot1 = buffer_debug[cur_index++];

			cur_pot1 ++;
		}

		p_node_1 ++;
	}

	// Second Edge potentials
	p_edge = edge_array + 1;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)		// loop over edges
	{
		for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
			for (cur_class_2 = 1; cur_class_2 <= num_class; cur_class_2 ++)
				p_edge->edge_pot[1][cur_class_1][cur_class_2] = p_edge->edge_pot[2][cur_class_1][cur_class_2] = 0;

		for (cur_class_1 = 1; cur_class_1 <= num_class; cur_class_1 ++)
		{
			if (target == 1)
				fscanf(fp_in, "%lg\n", &(p_edge->edge_pot[1][cur_class_1][cur_class_1]));
			else if (target == 2)
				p_edge->edge_pot[1][cur_class_1][cur_class_1] = buffer_debug[cur_index++];

			p_edge->edge_pot[2][cur_class_1][cur_class_1] = p_edge->edge_pot[1][cur_class_1][cur_class_1];
		}

		p_edge ++;
	}

	if (target == 1)
		fclose(fp_in);
}




/**
 *	Dump the inference results
 */
void CRFBasic::dump_infer_result(char * infer_file)
{
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	FILE *p_file_infer;
	Edge *p_edge;
	
	if(! infer_file)
		p_file_infer = fopen("infer_result.dat", "w");
	else
		p_file_infer = fopen(infer_file, "w");

	// First, dump node RB estimates
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		fprintf(p_file_infer, "Node %d:\t", cur_node);
		for (cur_class_1 = 1; cur_class_1 <= node_array[cur_node].num_class; cur_class_1 ++)
			fprintf(p_file_infer, "%g ", node_array[cur_node].rb_est[cur_class_1]);
		fprintf(p_file_infer, "\n");
	}

	// Second, dump edge RB estimates
	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		fprintf(p_file_infer, "Edge %d: %d -> %d \n", cur_edge, p_edge->p_node[1] - node_array, p_edge->p_node[2] - node_array);
		for (cur_class_1 = 1; cur_class_1 <= p_edge->p_node[1]->num_class; cur_class_1 ++)
		{
			for (cur_class_2 = 1; cur_class_2 <= p_edge->p_node[2]->num_class; cur_class_2 ++)
				fprintf(p_file_infer, "%g ", p_edge->rb_est[cur_class_1][cur_class_2]);
			fprintf(p_file_infer, "\n");
		}
		p_edge ++;
	}

	fclose(p_file_infer);
}


/**
 *	Output the graph and inference results to file
 */
void CRFBasic::dump_graph(char * spec_file, char * infer_file)
{
	dump_graph_specification(spec_file);
	dump_infer_result(infer_file);
}


/**
 *	Bits in selecter: (from rightmost)
 *	1.  Tree structure
 *	2.  Node conditionals
 *	4.  Node messages
 *	8.  Node sum_msg
 *	16. Samples drawn
 */
void CRFBasic::display_tree(Tree * cur_tree, unsigned selecter)
{
	int i, j, cur_node, cur_class, cur_son;
	double *pd, **pd2;
	Node *p_node;

	if(selecter & 1)	// 1. Tree structure
	{
		printf("Display Tree *****************\n");
		printf("num_node = %d\n", cur_tree->num_node);

		printf("Nodes are (global id):\n");
		for (i = 1; i <= cur_tree->num_node; i ++)
			printf("%d\t", cur_tree->local_to_global[i]);
		printf("\n");

		printf("Number of sons:\n");
		for (i = 1; i <= cur_tree->num_node; i ++)
			printf("%d\t", cur_tree->num_son[i]);
		printf("\n");

		printf("Messages to pass: %d\n", cur_tree->num_node - 1);
		for (i = 1; i <= cur_tree->num_node - 1; i ++)
		{
			printf("%d -> %d, with order_id = %d, tree_edge %d \n",
				cur_tree->send[i] - node_array, cur_tree->rec[i] - node_array, cur_tree->order_id[i], 
				cur_tree->tree_edge[i] - edge_array);
		}

		printf("Conditional Edges:\n");
		if(!cur_tree->cond_edge_array)	// for init_tree
			printf("NULL for Initial Tree.\n");
		else
		{
			for (i = 1; i <= cur_tree->num_node; i ++)
			{
				printf("%d) Node %d:\t", i, cur_tree->local_to_global[i]);
				for (j = cur_tree->cond_edge_index[i]; j < cur_tree->cond_edge_index[i+1]; j ++)
					printf("%d  ", cur_tree->cond_edge_array[j] - edge_array);
				printf("\n");
			}
		}
	}

	if(selecter & 16)		// print the current sample of the tree
	{
		printf("Sample of the tree:\n");
		for (cur_node = 1; cur_node <= cur_tree->num_node; cur_node ++)
		{
			p_node = node_array + cur_tree->local_to_global[cur_node];
			printf("%d -> %d\t", cur_tree->local_to_global[cur_node], p_node->sampled_value);
		}
		printf("\n");
	}

	for (cur_node = 1; cur_node <= cur_tree->num_node; cur_node ++)
	{
		p_node = node_array + cur_tree->local_to_global[cur_node];

		if (selecter & 14)
			printf("Node %d:\n", cur_tree->local_to_global[cur_node]);

		if (selecter & 2)		// 2. Node conditionals
		{
			printf("Conditionals:\n");
			pd = p_node->cond_pot + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
				printf("%g\t", *(pd ++));
			printf("\n");			
		}
		if (selecter & 4)		// 4. Node messages
		{
			printf("msg_buffer:\n");
			pd2 = p_node->msg_buffer + 1;
			for (cur_son = 1; cur_son <= p_node->num_son; cur_son ++)
			{
				pd = *(pd2 ++) + 1;
				for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
					printf("%g\t", *(pd ++));
				printf("\n");	
			}
		}
		if (selecter & 8)		// 8. Node sum_msg
		{
			printf("sum_msg:\n");
			pd = p_node->sum_msg + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
				printf("%g\t", *(pd ++));
			printf("\n");
		}
	}
	cin.get();
}


/**
 *	Display the potentials of all nodes and edges
 */
void CRFBasic::display_pots()
{
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	double *p_pot1, **p_pot2;
	Node *p_node_1, *p_node_2;
	Edge *p_edge;

	p_node_1 = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		printf("Node %d:\t", cur_node);
		p_pot1 = p_node_1->node_pot + 1;
		for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
			printf("%g\t", *(p_pot1 ++));
		p_node_1 ++;
		printf("\n");
	}

	p_edge = edge_array + 1;
	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		p_node_1 = p_edge->p_node[1];
		p_node_2 = p_edge->p_node[2];

		printf("Edge %d:  %d -> %d\n", cur_edge, p_edge->p_node[1] - node_array, p_edge->p_node[2] - node_array);
		
		p_pot2 = p_edge->edge_pot[1] + 1;

		for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
		{
			p_pot1 = *(p_pot2 ++) + 1;
			for (cur_class_2 = 1; cur_class_2 <= p_node_2->num_class; cur_class_2 ++)
			{
				printf("%g\t", *(p_pot1 ++));
			}

			printf("\n");
		}
		p_edge ++;
	}
}


/**
 *	Display the RB estimates of the graph, including nodes and edges
 */
void CRFBasic::display_RB()
{
	int cur_node, cur_edge, cur_class_1, cur_class_2;
	double *pd1, **pd2;
	Node *p_node_1, *p_node_2;
	Edge *p_edge;

	if (verbosity & 2)
	{
		printf("===========================================\n");
		// First, Node RBs
		printf("Node RB Estimates:\n");
		p_node_1 = node_array + 1;
		for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
		{
			printf("Node %d:\t", cur_node);
			pd1 = p_node_1->rb_est + 1;
			for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
				printf("%g\t", *(pd1 ++));
			p_node_1 ++;
			printf("\n");
		}
	}

	if(verbosity & 4)
	{
		// Second, Edge RBs
		printf("-------------------------------------------\n");
		printf("Edge RB Estimates:\n");
		p_edge = edge_array + 1;
		for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
		{
			p_node_1 = p_edge->p_node[1];
			p_node_2 = p_edge->p_node[2];
			printf("Edge %d:\t%d -> %d\n", cur_edge, p_node_1 - node_array, p_node_2 - node_array);

			pd2 = p_edge->rb_est + 1;
			for (cur_class_1 = 1; cur_class_1 <= p_node_1->num_class; cur_class_1 ++)
			{
				pd1 = *(pd2 ++) + 1;
				for (cur_class_2 = 1; cur_class_2 <= p_node_2->num_class; cur_class_2 ++)
					printf("%g\t", *(pd1 ++));
				printf("\n");
			}

			p_edge ++;
		}
	}
	cin.get();
}


/**
 *	The main file for testing
 */
int CRFBasic::main_test_tree(int *argc, char **argv)
{	
	init_tree_sampler(argc, argv);

	////////////////////////
	// These will be done repeatedly
	set_cur_obs_debug();
	set_node_fea_para_debug();
	set_edge_fea_para_debug();
	set_potentials();
	////////////////////////

	if (verbosity & 8)
		display_pots();		// for debug

	infer_graph();		// calculate the RB estimation

	cal_fea_expectation();

	return 0;
}


char * CRFBasic::get_string_max_param()
{
	return string_max_param;
}


int CRFBasic::get_num_params()
{
	return  global_num_node * (num_class * num_obs_dim + num_node_extra_fea)
				+ global_num_edge * num_edge_fea;
}


int CRFBasic::get_input_dim()
{
	return num_obs_dim * global_num_node;
}


void CRFBasic::load_sample_action(Vector & action)
{
	assert((int)action.size() == global_num_node);

	for(int cur_node = 0; cur_node < global_num_node; cur_node ++)
		action[cur_node] = node_array[cur_node + 1].sampled_value - 1;
}


void CRFBasic::fill_obs(Matrix & obs)
{
	int cur_node, cur_obs_dim;
	double **p_obs2, *p_obs1;

	// Set current observation
	p_obs2 = cur_observation + 1;
	for (cur_node = 0; cur_node < global_num_node; cur_node ++)
	{
		p_obs1 = *(p_obs2 ++) + 1;

		for (cur_obs_dim = 0; cur_obs_dim < num_obs_dim; cur_obs_dim ++)
			*(p_obs1 ++) = obs(cur_obs_dim, cur_node);		// cur_observation[cur_node][cur_obs_dim] =
	}
}


void CRFBasic::instant_step(Vector& cur_rewards) 
{
	if (use_local_reward)		// Local rewards
	{
		assert(cur_rewards.size() == (unsigned) global_num_node);
		for (int i = 1; i <= global_num_node; i ++)
			rewards[i] = cur_rewards[i - 1];
	}
	else
	{
		assert(cur_rewards.size() == 1);	// we just accept one global reward
		if (fabs(cur_rewards[0]) > 1e-5) 
			rewards[1] = cur_rewards[0];
	}	

	update_fea_para(rewards, 1);
}


void CRFBasic::batch_step() 
{
	if (use_local_reward)		// Local rewards
	{
		for (int i = 1; i <= global_num_node + global_num_edge; i ++)
			rewards[i] = 1;
	}
	else
	{
		rewards[1] = 1;
	}
	update_fea_para(rewards, 2);
}


void CRFBasic::reduce(Vector& v, int Stats)
{
	int cur_index, cur_item, cur_fea, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;

    assert(v.size() >= (size_t)(get_num_params()));

	cur_index = 0;

	if (Stats == 0)			// Approximator::PARAMS
	{
		//  First, nodes
		nfp3 = node_fea_para + 1;
		p_node = node_array + 1;
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			nfp2 = *(nfp3 ++) + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			{
				nfp1 = *(nfp2 ++) + 1;
				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
					v[cur_index ++] = *(nfp1 ++);
			}
			p_node ++;
		}

		//  Second, edges
		efp2 = edge_fea_para + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			efp1 = *(efp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				v[cur_index ++] = *(efp1 ++);
		} 	

		//  Third, extra node features
		if (num_node_extra_fea)
		{
			efp2 = node_extra_fea_para + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				efp1 = *(efp2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					v[cur_index ++] = *(efp1 ++);
			} 	
		}
	}
 	else if (Stats == 1)		// Approximator::GRADS (actually the expectation)
 	{
		//  First, nodes
		nfp3 = node_fea_exp + 1;
		p_node = node_array + 1;
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			nfp2 = *(nfp3 ++) + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			{
				nfp1 = *(nfp2 ++) + 1;
				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
					v[cur_index ++] = *(nfp1 ++);
			}
			p_node ++;
		}

		//  Second, edges
		efp2 = edge_fea_exp + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			efp1 = *(efp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				v[cur_index ++] = *(efp1 ++);
		}

		//  Third, extra node features
		if (num_node_extra_fea)
		{
			efp2 = node_extra_fea_exp + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				efp1 = *(efp2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					v[cur_index ++] = *(efp1 ++);
			}
		}
	}
	else if (Stats == 2)		// Approximator::TRACES
	{
		//  First, nodes
		nfp3 = node_fea_trace + 1;
		p_node = node_array + 1;
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			nfp2 = *(nfp3 ++) + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			{
				nfp1 = *(nfp2 ++) + 1;
				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
					v[cur_index ++] += *(nfp1 ++);
			}
			p_node ++;
		}

		//  Second, edges
		efp2 = edge_fea_trace + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			efp1 = *(efp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				v[cur_index ++] += *(efp1 ++);
		}

		//  Third, extra node features
		if (num_node_extra_fea)
		{
			efp2 = node_extra_fea_trace + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				efp1 = *(efp2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					v[cur_index ++] += *(efp1 ++);
			}
		}
	}
	else
		NYI	
}


void CRFBasic::scatter(Vector& v, int Stats)
{
	int cur_index, cur_item, cur_fea, cur_class, cur_obs_dim;
	double ***nfp3, **nfp2, *nfp1, **efp2, *efp1;
	Node *p_node;

    assert(v.size() == (size_t)(get_num_params()));

	cur_index = 0;
 	
	if (Stats == 0) 		// Approximator::PARAMS
	{
		//  First, nodes
		nfp3 = node_fea_para + 1;
		p_node = node_array + 1;
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			nfp2 = *(nfp3 ++) + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			{
				nfp1 = *(nfp2 ++) + 1;
				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
					*(nfp1 ++) = v[cur_index ++];
			}
			p_node ++;
		}

		//  Second, edges
		efp2 = edge_fea_para + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			efp1 = *(efp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				*(efp1 ++) = v[cur_index ++];
		} 	 	

		// Third, extra node features
		if (num_node_extra_fea)
		{
			efp2 = node_extra_fea_para + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				efp1 = *(efp2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					*(efp1 ++) = v[cur_index ++];
			} 	 	
		}
	}
 	else if (Stats == 1)		// Approximator::GRADS (actually the expectation)
 	{
		//  First, nodes
		nfp3 = node_fea_exp + 1;
		p_node = node_array + 1;
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			nfp2 = *(nfp3 ++) + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			{
				nfp1 = *(nfp2 ++) + 1;
				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
					*(nfp1 ++) = v[cur_index ++];
			}
			p_node ++;
		}

		//  Second, edges
		efp2 = edge_fea_exp + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			efp1 = *(efp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				*(efp1 ++) = v[cur_index ++];
		}

			// Third, extra node features
		if (num_node_extra_fea)
		{
			efp2 = node_extra_fea_exp + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				efp1 = *(efp2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					*(efp1 ++) = v[cur_index ++];
			} 	
		}
	}
	else if (Stats == 2)		// Approximator::TRACES
	{
		//  First, nodes
		nfp3 = node_fea_trace + 1;
		p_node = node_array + 1;
		for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
		{
			nfp2 = *(nfp3 ++) + 1;
			for (cur_class = 1; cur_class <= p_node->num_class; cur_class ++)
			{
				nfp1 = *(nfp2 ++) + 1;
				for (cur_obs_dim = 1; cur_obs_dim <= num_obs_dim; cur_obs_dim ++)
					*(nfp1 ++) = v[cur_index ++];
			}
			p_node ++;
		}

		//  Second, edges
		efp2 = edge_fea_trace + 1;
		for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
		{
			efp1 = *(efp2 ++) + 1;
			for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
				*(efp1 ++) = v[cur_index ++];
		}

		if (num_node_extra_fea)
		{
			// Third, extra node features
			efp2 = node_extra_fea_trace + 1;
			for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
			{
				efp1 = *(efp2 ++) + 1;
				for (cur_fea = 1; cur_fea <= num_node_extra_fea; cur_fea ++)
					*(efp1 ++) = v[cur_index ++];
			} 	
		}
	}
	else
		NYI
}


}
