#ifndef CRFBasic_hh
#define CRFBasic_hh

namespace tree_mcmc {

//#define RB_EDGE_SINGLE
//#define RB_NODE_SINGLE

#define NINFINITY -1e307

typedef struct Node
{
	int		sampled_value;	// The sampled value of this node in the last iteration
	int		num_class;		// How many classes/values can this node assume
	int		num_son;		// How many sons this node has
	int		max_num_son;	// Since Node space is shared by different trees, we need 
							//  the maximum possible number of sons in order to allocate space	
	double	*node_pot;		// Link to the node potential list
	double	*cond_pot;		// Link to the conditional potential list (has already multiplied by node potential)
	double	**msg_buffer;	// The buffer for all messages it has received [1][2]: [1] son_id, [2] label
	double	*sum_msg;		// sum of messages (for cache purposes)
	double	*rb_est;		// The RB estimation of the marginal probability
} Node;


typedef struct Edge
{
	Node *p_node[3];			// Link to the two end nodes ptr1 < ptr2
	double **edge_pot[3];		// The edge potential tables (essentially the same). 
								//   [1] for i (row) < j (col), and [2] for i (row) > j (col)
	double **rb_est;			// The R-B estimation of edge potentials
} Edge;


typedef struct Tree 
{
	/* 
	 *	Backward (sender, receiver) (forward just reverse order)
	 *	This forms a chain (sender, receiver) for forward, 
	 *   and a chain (sender, receiver) for backward (just reverse the order). 
	 */
	Node	**send, **rec;		
	int		*order_id;				// Backward the rank of son among sons (forward just reverse order)
	Edge	**tree_edge;			// Backward edge link (forward just reverse order)
	int		num_node;				// Number of nodes in this tree (forward just reverse order)
	int		*local_to_global;		// The id of the nodes in this tree
	int		*num_son;				// How many sons does each node has
	Edge	**cond_edge_array;		// for conditional use
	int		*cond_edge_index;		// [i] is the i^th node's starting conditional edge index
	bool	*cal_edge_rb;			// [i] denotes whether the i^th edge will contribute to the edge RB estimation.
	bool	cal_node_rb;			// whether to calculate the node RB estimates for this tree
} Tree;


typedef struct Partition
{
	int num_trees;		// How many trees are there in the partition
	Tree *trees;	// The list of trees
} Partition;


class CRFBasic
{
protected:
	int global_num_node;		// Number of nodes (also number of agents for graph, not necessarily # agent of the original application)
	int global_num_edge;		// Number of edges
	int global_num_partition;	// Number of partitions
	int num_class;				// number of classes, only for grid

	int **adj_matrix;			// Adjacency matrix

	Node *node_array;			// array of the nodes
	Edge *edge_array;			// array of the edges

	Partition *partitions;		// all the partitions (all sets of trees)
	Tree *init_tree;			// the tree for initialization
	double * sample_buf;		// sample buffer shared by all nodes
	double ** rb_buf;			// the buffer for edge RB.  Node RB uses sample_buf.

	double maxNodeRand;			// maximum random range of node feature parameters
	double maxEdgeRand;			// maximum random range of edge feature parameters

	// Input parameters
	int		num_obs_dim;				// number of node observation dimension
	int		num_node_extra_fea;			// number of extra node features
	int		num_edge_fea;				// number of edge features

	double step_size[4];

	char string_max_param[1000];

	/**	
	 *	Current parameters for the node features, [1][2][3]
	 *		First index: node id;	second index: label;	third index: observation dimension
	 */
	double	***node_fea_para;


	/**
	 *	Current parameters for the extra (non_NN) node features, [1][2]
	 *		First index: node id;	second index: extra feature index
	 */
	double	**node_extra_fea_para;

	/**	
	 *	Current parameters for the edge features, [1][2]
	 *		First index: edge id;	second index: feature index
	 */
	double	**edge_fea_para;			

	/**	
	 *	The expectation of node features, [1][2].  1st index: node id,  2nd index: input dim id
	 */
	double	***node_fea_exp;

	/**	
	 *	The expectation of node extra features, [1][2].  1st index: node id,  2nd index: feature index
	 */
	double	**node_extra_fea_exp;

	/**	
	 *	The expectation of edge features [1][2].  First index: edge id;	second index: feature index
	 */
	double	**edge_fea_exp;

	/**	
	 *	The trace of the node features, [1][2][3]
	 *		First index: node id;	second index: label;	third index: observation/feature dimension
	 */
	double	***node_fea_trace;			

	/**	
	 *	The trace of the node extra features, [1][2]
	 *		First index: node id;	second index: feature index
	 */
	double	**node_extra_fea_trace;

	/**	
	 *	The trace of the edge features, [1][2]
	 *		First index: edge id;	second index: feature index
	 */
	double	**edge_fea_trace;			

	/**	
	 *	The current observation matrix, [1][2]
	 *		First index: node id;	second index: observation dimension index
	 */
	double	**cur_observation;	

	// The real value of edge features for CRF
	double edge_fea_intensity;

	int edge_fea_sel;

	/**
	 *	Control of verbosity level (output of intermediate results)
	 *	1.  Display all the trees and init_tree
	 *	2.  Display the RB estimates of nodes after all iterations of tree MCMC sampling
	 *	4.  Display the RB estimates of edges after all iterations of tree MCMC sampling
	 *	8.  Display the node and edge potentials at the outset
	 *	16. Display number of iterations if max_time is used
	 */
	int verbosity;

	int cur_step;
	int cycleTime;
	int stepsPerControl;
	double bias;

	bool use_local_reward;
	int edge_reward_def;

	double *rewards;

	char message[200];
	double *buffer_debug;
	int *last_action;


protected:		
	
	// member functions

	//////////////////////////////////////////////////////////////////////////
	// Virtual functions
	//////////////////////////////////////////////////////////////////////////

	virtual void alloc_adj_matrix() = 0;
	virtual void associate_edge_with_nodes() = 0;
	virtual int get_edge_id(int node1, int node2) = 0;



	//////////////////////////////////////////////////////////////////////////
	// Concrete Functions
	//////////////////////////////////////////////////////////////////////////


	/**
	 *	Initialize the data structure of a tree.
	 */
	void build_tree(Tree *cur_tree, int root_id = -1);


	/**
	 *	Clear the RB estimations
	 */
	void clear_RB();


	/**
	 *	Normalize the RB estimations
	 */
	void normalize_RB();


	/**
	 *	If you DO want RB, then use infer_distribution
	 *	Use sample_distribution iff you do NOT need RB.
	 *	Sample from exp(log_prob[]), which has num_class elements
	 *	Just sample, NO calculation of the RB estimation and store in sample_buf[]
	 *	It Subtracts the maximum element.
	 */
	int sample_distribution(double *log_prob, int num_class);


	/**
	 *	If you do NOT want RB, then use sample_distribution
	 *	Use infer_distribution iff you need RB.
	 *	Sample from exp(log_prob[]), which has num_class elements
	 *	Also calculate the RB estimation and store in !!!sample_buf[]!!!
	 *	It does NOT subtract the maximum element.
	 */
	int infer_distribution(double *log_prob, int num_class);


	/**
	 *	Calculate the conditional potentials of cur_tree
	 */
	void cal_cond_pot(Tree *cur_tree);


	/**
	 *	Perform forward message passing (from leaves to root)
	 */
	void forward_msg_pass(Tree *cur_tree);


	/**
	 *	Get a sample of the root and calculate the node marginal. 
	 *	Run between forward and backward msg prop
	 */
	void infer_root(Tree *cur_tree);


	/**
	 *	Get a sample of the root, run between forward and backward msg prop
	 */
	void sample_root(Tree *cur_tree);


	/**
	 *	Calculate the RB estimation of edges.
	 *	No need to take care of subtracting the maximum element of any ORIGINAL data structure
	 *	Orientation:  1 if large -> small; 2 if small -> large
	 */
	void fill_edge_rb(Edge *p_edge, Node *p_send, Node *p_rec, int order_id, int orient);


	/**
	 *	Calculate the RB estimation of nodes.
	 *	Potentials are stored in p_node->sum_msg[]
	 *	!!! NOT !!! internally subtract the largest element in p_node->sum_msg[]
	 *	This must be taken care of externally.
	 */
	void fill_node_rb(Node *p_node);


	/**
	 *	Perform backward message passing, AND sampling (from leaves to root)
	 */
	void backward_msg_pass(Tree *cur_tree);


	/**
	 *	Perform backward sampling, NO backward msg passing (from leaves to root)
	 *	This is much simpler than backward_msg_pass
	 */
	void backward_sample(Tree *cur_tree);


	/**
	 *	Draw a sample of tree
	 *	Almost only used for init_tree
	 */
	void sample_tree(Tree *cur_tree);


	/**
	 *	Return the node array point in order to read the sampling result
	 */	
	Node * get_sample_pointer();


	/**
	 *	allocate space for cur_observation, node_fea_para, edge_fea_para, node_fea_grad, edge_fea_grad
	 */
	void alloc_fea();


	/**
	 *	Allocate the space of edges and nodes for running MCMC
	 */
	void alloc_potentials();


	/**
	 *	Randomly initialize the edge_fea_para[1][2]
	 *			First index: edge id;	second index: feature index
	 *	For debug only.
	 */
	void set_cur_obs_debug();


	/**
	 *	Fill up the node_fea_para[1][2][3]
	 *			First index: node id;	second index: label;	third index: observation 
	 *	For debug only.
	 */
	void set_node_fea_para_debug();


	/**
	 *	Fill up the edge_fea_para[1][2]
	 *			First index: edge id;	second index: feature index
	 *	For debug only.
	 */
	void set_edge_fea_para_debug();

	/**
	 *	Fill up the node/edge potentials.
	 *	For debug only.
	 */
	void set_potentials_debug();

	/**
	 *	Calculate the node potentials.
	 */
	void set_node_potentials();

	// Write the potentials to file or buffer_debug, for debug purpose
	// target = 1: write to file
	// target = 2: write to buffer_debug
	void write_potentials(int target = 1);

	// Read the potentials from file or buffer_debug, for debug purpose
	// target = 1: read from file
	// target = 2: read from buffer_debug
	void read_potentials(int target = 1);

	/************************************************************************/
	/*            Display functions for debugging                           */
	/************************************************************************/

	/**
	 *	Dump the graph parameters
	 */
	void dump_graph_specification(char * spec_file);


	/**
	 *	Dump the inference results
	 */
	void dump_infer_result(char * infer_file);


	/**
	 *	Output the graph and inference results to file
	 */
	void dump_graph(char * spec_file, char * infer_file);


	/**
	 *	Bits in selecter: (from rightmost);
	 *	1.  Tree structure
	 *	2.  Node conditionals
	 *	4.  Node messages
	 *	8.  Node sum_msg
	 *	16. Samples drawn
	 */
	void display_tree(Tree * cur_tree, unsigned selecter);


	/**
	 *	Display the potentials of all nodes and edges
	 */
	void display_pots();


	/**
	 *	Display the RB estimates of the graph, including nodes and edges
	 */
	void display_RB();


	/**
	 *	The main file for testing
	 */
	int main_test_tree(int *argc, char **argv);

	virtual void init_graph(int *argc, char **argv) = 0;

public:

	CRFBasic(int *argc, char** argv);
	
	virtual void init_tree_sampler(int *argc, char **argv);

	/**
	 *	Get the maximum feature parameter
	 */
	double get_max_param(); 

	/**
	 *	Randomly initialize the node_fea_para[1][2][3]
	 *			First index: node id;	second index: label;	third index: observation
	 */
	void randomize_node_fea_para(double range = -1);


	/**
	 *	Randomly initialize the edge_fea_para[1][2]
	 *			First index: edge id;	second index: feature index
	 */
	void randomize_edge_fea_para(double range = -1);

	/**
	 *	Reset the trace elements to 0
	 */
	void reset_trace(); 


	/**
	 *	Discount the trace by factor discount
	 */
	void discount_trace(double discount); 


	/**
	 *	Re-initialize randomly the feature parameters
	 */
	void reset_params(); 


	/**
	 *	Get the string_max_param
	 */
	char * get_string_max_param();
	
	/**
	 *	Calculate the node feature expectations.
	 */
	void cal_node_fea_exp();


	/**
	 *	Calculate the node feature traces.
	 */
	void fill_node_trace();

	void self_test();

	
	/**
	 *	Set the step size of CRF
	 */
	void set_step_size(double *new_step_size);

	int get_cur_step() {return cur_step;};

	virtual void graph_para_write(std::ostream & o);
	void graph_para_read(std::istream & o);

	void update_fea_para(double *reward, int source);
	int get_num_params();
	int get_input_dim();

	void load_sample_action(Vector & action);
	void fill_obs(Matrix & obs);

	void instant_step(Vector& cur_rewards);
	void batch_step();

	void scatter(Vector& v, int Stats);
	void reduce(Vector& v, int Stats);

	int get_num_edge() {return global_num_edge;};
	int get_num_agent() {return global_num_node;};
	int get_num_node_extra_fea() {return num_node_extra_fea;};
	int get_num_class() {return num_class;};
	int get_num_edge_fea() {return num_edge_fea;};

	void incre_step() {cur_step ++;};

	virtual void load_last_action() {;};

	// Abstract
	virtual void sample_graph() = 0;
	virtual void infer_graph() = 0;

	virtual void cal_fea_expectation() = 0;
	virtual void fill_trace() = 0;
	virtual void set_potentials() = 0;
	virtual void cal_diagnosis() {;};
	virtual void buffer_last_action(int * actions) {;};
	virtual char * get_diagnosis(int argc, char ** argv) {return string_max_param;};
	virtual void fill_edge_reward(double *reward) = 0;

	virtual bool distribute_reward() {return use_local_reward;};
};

}
#endif
